#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
IDAES Moving Bed Model.

The moving bed model is a 1D axially discretized model with a gas and
solid phase and a counter-current flow direction. The model captures
the gas-solid interaction between both phases through reaction, mass
and heat transfer.

Equations written in this model were derived from:
A. Ostace, A. Lee, C.O. Okoli, A.P. Burgard, D.C. Miller, D. Bhattacharyya,
Mathematical modeling of a moving-bed reactor for chemical looping combustion
of methane, in: M.R. Eden, M. Ierapetritou, G.P. Towler (Eds.),13th Int. Symp.
Process Syst. Eng. (PSE 2018), Computer-Aided Chemical Engineering 2018,
pp. 325–330 , San Diego, CA.

Assumptions:
Property package contains temperature and pressure variables.
Property package contains minimum fluidization velocity.

"""
from __future__ import division

# Import Python libraries
import matplotlib.pyplot as plt

# Import Pyomo libraries
from pyomo.environ import (Var, Param, Reals, value, SolverFactory,
                           TransformationFactory, Constraint,
                           TerminationCondition)
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.dae import ContinuousSet

# Import IDAES cores
from idaes.core import (ControlVolume1DBlock,
                        UnitModelBlockData,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        FlowDirection)
from idaes.core.util.config import (is_physical_parameter_block,
                                    is_reaction_parameter_block)
from idaes.core.util.exceptions import (ConfigurationError,
                                        BurntToast)
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.control_volume1d import DistributedVars
from idaes.core.util.constants import Constants as constants
from idaes.core.util.math import smooth_abs
import idaes.logger as idaeslog
from idaes.core.util import get_solver

__author__ = "Chinedu Okoli", "Anca Ostace"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("MBR")
class MBRData(UnitModelBlockData):
    """Standard Moving Bed Unit Model Class."""

    # Create template for unit level config arguments
    CONFIG = UnitModelBlockData.CONFIG()

    # Unit level config arguments
    CONFIG.declare("finite_elements", ConfigValue(
        default=10,
        domain=int,
        description="Number of finite elements length domain",
        doc="""Number of finite elements to use when discretizing length
domain (default=20)"""))
    CONFIG.declare("length_domain_set", ConfigValue(
        default=[0.0, 1.0],
        domain=list,
        description="Number of finite elements length domain",
        doc="""length_domain_set - (optional) list of point to use to
initialize a new ContinuousSet if length_domain is not
provided (default = [0.0, 1.0])"""))
    CONFIG.declare("transformation_method", ConfigValue(
        default="dae.finite_difference",
        description="Method to use for DAE transformation",
        doc="""Method to use to transform domain. Must be a method recognised
by the Pyomo TransformationFactory,
**default** - "dae.finite_difference".
**Valid values:** {
**"dae.finite_difference"** - Use a finite difference transformation method,
**"dae.collocation"** - use a collocation transformation method}"""))
    CONFIG.declare("transformation_scheme", ConfigValue(
        default=None,
        domain=In([None, "BACKWARD", "FORWARD", "LAGRANGE-RADAU"]),
        description="Scheme to use for DAE transformation",
        doc="""Scheme to use when transforming domain. See Pyomo
documentation for supported schemes,
**default** - None.
**Valid values:** {
**None** - defaults to "BACKWARD" for finite difference transformation method,
and to "LAGRANGE-RADAU" for collocation transformation method,
**"BACKWARD"** - Use a finite difference transformation method,
**"FORWARD""** - use a finite difference transformation method,
**"LAGRANGE-RADAU""** - use a collocation transformation method}"""))
    CONFIG.declare("collocation_points", ConfigValue(
        default=3,
        domain=int,
        description="Number of collocation points per finite element",
        doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)"""))
    CONFIG.declare("flow_type", ConfigValue(
        default="counter_current",
        domain=In(['counter_current']),
        description="Flow configuration of Moving Bed",
        doc="""Flow configuration of Moving Bed
- counter_current: gas side flows from 0 to 1
solid side flows from 1 to 0"""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.componentTotal,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentTotal.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.enthalpyTotal,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("pressure_drop_type", ConfigValue(
        default="simple_correlation",
        domain=In(["simple_correlation", "ergun_correlation"]),
        description="Construction flag for type of pressure drop",
        doc="""Indicates what type of pressure drop correlation should be used,
**default** - "simple_correlation".
**Valid values:** {
**"simple_correlation"** - Use a simplified pressure drop correlation,
**"ergun_correlation"** - Use the ergun equation.}"""))

    # Create template for phase specific config arguments
    _PhaseTemplate = UnitModelBlockData.CONFIG()
    _PhaseTemplate.declare("has_equilibrium_reactions", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Equilibrium reaction construction flag",
        doc="""Indicates whether terms for equilibrium controlled reactions
should be constructed,
**default** - True.
**Valid values:** {
**True** - include equilibrium reaction terms,
**False** - exclude equilibrium reaction terms.}"""))
    _PhaseTemplate.declare("property_package", ConfigValue(
        default=None,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object"""))
    _PhaseTemplate.declare("property_package_args", ConfigValue(
        default={},
        domain=dict,
        description="Arguments for constructing gas property package",
        doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)"""))
    _PhaseTemplate.declare("reaction_package", ConfigValue(
        default=None,
        domain=is_reaction_parameter_block,
        description="Reaction package to use for control volume",
        doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}"""))
    _PhaseTemplate.declare("reaction_package_args", ConfigBlock(
        implicit=True,
        implicit_domain=ConfigBlock,
        description="Arguments to use for constructing reaction packages",
        doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}"""))

    # Create individual config blocks for gas and solid sides
    CONFIG.declare("gas_phase_config",
                   _PhaseTemplate(doc="gas phase config arguments"))
    CONFIG.declare("solid_phase_config",
                   _PhaseTemplate(doc="solid phase config arguments"))

    # =========================================================================
    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to build default attributes
        super(MBRData, self).build()

        # Consistency check for transformation method and transformation scheme
        if (self.config.transformation_method == "dae.finite_difference" and
                self.config.transformation_scheme is None):
            self.config.transformation_scheme = "BACKWARD"
        elif (self.config.transformation_method == "dae.collocation" and
                self.config.transformation_scheme is None):
            self.config.transformation_scheme = "LAGRANGE-RADAU"
        elif (self.config.transformation_method == "dae.finite_difference" and
              self.config.transformation_scheme != "BACKWARD" and
              self.config.transformation_scheme != "FORWARD"):
            raise ConfigurationError("{} invalid value for "
                                     "transformation_scheme argument. "
                                     "Must be ""BACKWARD"" or ""FORWARD"" "
                                     "if transformation_method is"
                                     " ""dae.finite_difference""."
                                     .format(self.name))
        elif (self.config.transformation_method == "dae.collocation" and
                self.config.transformation_scheme != "LAGRANGE-RADAU"):
            raise ConfigurationError("{} invalid value for "
                                     "transformation_scheme argument."
                                     "Must be ""LAGRANGE-RADAU"" if "
                                     "transformation_method is"
                                     " ""dae.collocation""."
                                     .format(self.name))

        # Set flow directions for the control volume blocks
        # Gas flows from 0 to 1, solid flows from 1 to 0
        # An if statement is used here despite only one option to allow for
        # future extensions to other flow configurations
        if self.config.flow_type == "counter_current":
            set_direction_gas = FlowDirection.forward
            set_direction_solid = FlowDirection.backward
        else:
            raise BurntToast(
                    "{} encountered unrecognized argument "
                    "for flow type. Please contact the IDAES"
                    " developers with this bug.".format(self.name))

        # Set arguments for gas sides if homoogeneous reaction block
        if self.config.gas_phase_config.reaction_package is not None:
            has_rate_reaction_gas_phase = True
        else:
            has_rate_reaction_gas_phase = False

        # Set arguments for gas and solid sides if heterogeneous reaction block
        if self.config.solid_phase_config.reaction_package is not None:
            has_rate_reaction_solid_phase = True
            has_mass_transfer_gas_phase = True
        else:
            has_rate_reaction_solid_phase = False
            has_mass_transfer_gas_phase = False

        # Set heat transfer terms
        if self.config.energy_balance_type != EnergyBalanceType.none:
            has_heat_transfer = True
        else:
            has_heat_transfer = False

        # Set heat of reaction terms
        if (self.config.energy_balance_type != EnergyBalanceType.none
                and self.config.gas_phase_config.reaction_package is not None):
            has_heat_of_reaction_gas_phase = True
        else:
            has_heat_of_reaction_gas_phase = False

        if (self.config.energy_balance_type != EnergyBalanceType.none
                and self.config.solid_phase_config.
                reaction_package is not None):
            has_heat_of_reaction_solid_phase = True
        else:
            has_heat_of_reaction_solid_phase = False

        # Create a unit model length domain
        self.length_domain = ContinuousSet(
                                bounds=(0.0, 1.0),
                                initialize=self.config.length_domain_set,
                                doc="Normalized length domain")

    # =========================================================================
        """ Build Control volume 1D for gas phase and
            populate gas control volume"""

        self.gas_phase = ControlVolume1DBlock(default={
            "transformation_method": self.config.transformation_method,
            "transformation_scheme": self.config.transformation_scheme,
            "finite_elements": self.config.finite_elements,
            "collocation_points": self.config.collocation_points,
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "area_definition": DistributedVars.variant,
            "property_package": self.config.gas_phase_config.property_package,
            "property_package_args":
                self.config.gas_phase_config.property_package_args,
            "reaction_package": self.config.gas_phase_config.reaction_package,
            "reaction_package_args":
                self.config.gas_phase_config.reaction_package_args})

        self.gas_phase.add_geometry(
                length_domain=self.length_domain,
                length_domain_set=self.config.length_domain_set,
                flow_direction=set_direction_gas)

        self.gas_phase.add_state_blocks(
            information_flow=set_direction_gas,
            has_phase_equilibrium=False)

        if self.config.gas_phase_config.reaction_package is not None:
            self.gas_phase.add_reaction_blocks(
                    has_equilibrium=self.config.gas_phase_config.
                    has_equilibrium_reactions)

        self.gas_phase.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=False,
            has_mass_transfer=has_mass_transfer_gas_phase,
            has_rate_reactions=has_rate_reaction_gas_phase)

        self.gas_phase.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=has_heat_transfer,
            has_heat_of_reaction=has_heat_of_reaction_gas_phase)

        self.gas_phase.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

    # =========================================================================
        """ Build Control volume 1D for solid phase and
            populate solid control volume"""

        # Set argument for heterogeneous reaction block
        self.solid_phase = ControlVolume1DBlock(default={
            "transformation_method": self.config.transformation_method,
            "transformation_scheme": self.config.transformation_scheme,
            "finite_elements": self.config.finite_elements,
            "collocation_points": self.config.collocation_points,
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "area_definition": DistributedVars.variant,
            "property_package":
                self.config.solid_phase_config.property_package,
            "property_package_args":
                self.config.solid_phase_config.property_package_args,
            "reaction_package":
                self.config.solid_phase_config.reaction_package,
            "reaction_package_args":
                self.config.solid_phase_config.reaction_package_args})

        self.solid_phase.add_geometry(
                length_domain=self.length_domain,
                length_domain_set=self.config.length_domain_set,
                flow_direction=set_direction_solid)

        self.solid_phase.add_state_blocks(
            information_flow=set_direction_solid,
            has_phase_equilibrium=False)

        if self.config.solid_phase_config.reaction_package is not None:
            # TODO - a generalization of the heterogeneous reaction block
            # The heterogeneous reaction block does not use the
            # add_reaction_blocks in control volumes as control volumes are
            # currently setup to handle only homogeneous reaction properties.
            # Thus appending the heterogeneous reaction block to the
            # solid state block is currently hard coded here.

            tmp_dict = dict(**self.config.solid_phase_config.
                            reaction_package_args)
            tmp_dict["gas_state_block"] = self.gas_phase.properties
            tmp_dict["solid_state_block"] = self.solid_phase.properties
            tmp_dict["has_equilibrium"] = (self.config.solid_phase_config.
                                           has_equilibrium_reactions)
            tmp_dict["parameters"] = (self.config.solid_phase_config.
                                      reaction_package)
            self.solid_phase.reactions = (
                    self.config.solid_phase_config.reaction_package.
                    reaction_block_class(
                        self.flowsheet().time,
                        self.length_domain,
                        doc="Reaction properties in control volume",
                        default=tmp_dict))

        self.solid_phase.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=False,
            has_mass_transfer=False,
            has_rate_reactions=has_rate_reaction_solid_phase)

        self.solid_phase.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=has_heat_transfer,
            has_heat_of_reaction=has_heat_of_reaction_solid_phase)

        self.solid_phase.add_momentum_balances(
            balance_type=MomentumBalanceType.none,
            has_pressure_change=False)

    # =========================================================================
        """ Add ports"""
        # Add Ports for gas side
        self.add_inlet_port(name="gas_inlet", block=self.gas_phase)
        self.add_outlet_port(name="gas_outlet", block=self.gas_phase)

        # Add Ports for solid side
        self.add_inlet_port(name="solid_inlet", block=self.solid_phase)
        self.add_outlet_port(name="solid_outlet", block=self.solid_phase)

    # =========================================================================
        """ Add performace equation method"""
        self._apply_transformation()
        self._make_performance()

    # =========================================================================
    def _apply_transformation(self):
        """
        Method to apply DAE transformation to the Control Volume length domain.
        Transformation applied will be based on the Control Volume
        configuration arguments.
        """
        if self.config.finite_elements is None:
            raise ConfigurationError(
                    "{} was not provided a value for the finite_elements"
                    " configuration argument. Please provide a valid value."
                    .format(self.name))

        if self.config.transformation_method == "dae.finite_difference":
            self.discretizer = TransformationFactory(
                                    self.config.transformation_method)
            self.discretizer.apply_to(self,
                                      wrt=self.length_domain,
                                      nfe=self.config.finite_elements,
                                      scheme=self.config.transformation_scheme)
        elif self.config.transformation_method == "dae.collocation":
            self.discretizer = TransformationFactory(
                                    self.config.transformation_method)
            self.discretizer.apply_to(
                self,
                wrt=self.length_domain,
                nfe=self.config.finite_elements,
                ncp=self.config.collocation_points,
                scheme=self.config.transformation_scheme)

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        # local aliases used to shorten object names
        gas_phase = self.config.gas_phase_config
        solid_phase = self.config.solid_phase_config

        # Declare Mutable Parameters
        self.eps = Param(mutable=True,
                         default=1e-8,
                         doc='Smoothing Factor for Smooth IF Statements')

        # Unit Model variables
        self.bed_diameter = Var(domain=Reals,
                                initialize=1,
                                doc='Reactor diameter [m]')
        self.bed_area = Var(domain=Reals,
                            initialize=1,
                            doc='Reactor cross-sectional area [m2]')
        self.bed_height = Var(domain=Reals, initialize=1,
                              doc='Bed length [m]')

        # Phase specific variables
        self.velocity_superficial_gas = Var(
                self.flowsheet().time,
                self.length_domain,
                domain=Reals, initialize=0.05,
                doc='Gas superficial velocity [m/s]')
        self.velocity_superficial_solid = Var(
                self.flowsheet().time,
                domain=Reals, initialize=0.005,
                doc='Solid superficial velocity [m/s]')

        # Dimensionless numbers, mass and heat transfer coefficients
        self.Re_particle = Var(self.flowsheet().time,
                               self.length_domain,
                               domain=Reals, initialize=1.0,
                               doc='Particle Reynolds number [-]')

        self.Pr = Var(self.flowsheet().time,
                      self.length_domain,
                      domain=Reals, initialize=1.0,
                      doc='Prandtl number of gas in bed [-]')

        self.Nu_particle = Var(self.flowsheet().time,
                               self.length_domain,
                               domain=Reals, initialize=1.0,
                               doc='Particle Nusselt number [-]')
        self.gas_solid_htc = Var(self.flowsheet().time,
                                 self.length_domain,
                                 domain=Reals, initialize=1.0,
                                 doc='Gas-solid heat transfer coefficient'
                                 '[kJ/(m2Ks)]')

        # Fixed variables (these are parameters that can be estimated)
        self.bed_voidage = Var(domain=Reals,
                               initialize=0.4,
                               doc="Bed voidage [-]")
        self.bed_voidage.fix()

    # =========================================================================
        # Add performance equations

        # ---------------------------------------------------------------------
        # Geometry contraints

        # Bed area
        @self.Constraint(doc="Bed area")
        def bed_area_eqn(b):
            return b.bed_area == (
                    constants.pi*(0.5*b.bed_diameter)**2)

        # Area of gas side, and solid side
        @self.Constraint(self.flowsheet().time,
                         self.length_domain,
                         doc="Gas side area")
        def gas_phase_area(b, t, x):
            return (b.gas_phase.area[t, x] ==
                    b.bed_area*b.bed_voidage)

        @self.Constraint(self.flowsheet().time,
                         self.length_domain,
                         doc="Solid side area")
        def solid_phase_area(b, t, x):
            return (b.solid_phase.area[t, x] ==
                    b.bed_area*(1-b.bed_voidage))

        # Length of gas side, and solid side
        @self.Constraint(doc="Gas side length")
        def gas_phase_length(b):
            return (b.gas_phase.length == b.bed_height)

        @self.Constraint(doc="Solid side length")
        def solid_phase_length(b):
            return (b.solid_phase.length == b.bed_height)

        # ---------------------------------------------------------------------
        # Hydrodynamic contraints

        # Gas superficial velocity
        @self.Constraint(self.flowsheet().time,
                         self.length_domain,
                         doc="Gas superficial velocity")
        def gas_super_vel(b, t, x):
            return (b.velocity_superficial_gas[t, x] * b.bed_area *
                    b.gas_phase.properties[t, x].dens_mol ==
                    b.gas_phase.properties[t, x].flow_mol)

        # Solid superficial velocity
        @self.Constraint(self.flowsheet().time,
                         self.length_domain,
                         doc="Solid superficial velocity")
        # This equation uses inlet values to compute the constant solid
        # superficial velocity, and then computes the solid particle density
        # through the rest of the bed.
        def solid_super_vel(b, t, x):
            return (b.velocity_superficial_solid[t] * b.bed_area *
                    b.solid_phase.properties[t, x].dens_mass_particle ==
                    b.solid_phase.properties[t, x].flow_mass)

        # Gas side pressure drop calculation
        if (self.config.has_pressure_change and
            self.config.pressure_drop_type ==
                "simple_correlation"):
            # Simplified pressure drop
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Gas side pressure drop calculation -"
                                 "simplified pressure drop")
            def gas_phase_config_pressure_drop(b, t, x):
                return b.gas_phase.deltaP[t, x]*1e5 == -0.2*(
                        b.velocity_superficial_gas[t, x] *
                        (b.solid_phase.properties[t, x].dens_mass_particle -
                         b.gas_phase.properties[t, x].dens_mass))
        elif (self.config.has_pressure_change and
              self.config.pressure_drop_type == "ergun_correlation"):
            # Ergun equation
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Gas side pressure drop calculation -"
                                 "ergun equation")
            def gas_phase_config_pressure_drop(b, t, x):
                return (1e2*-b.gas_phase.deltaP[t, x]*1e5 ==
                        1e2*(
                        150*(1 - b.bed_voidage) ** 2 *
                        b.gas_phase.properties[t, x].visc_d *
                        (b.velocity_superficial_gas[t, x] +
                         b.velocity_superficial_solid[t]) /
                        (b.solid_phase.properties[t, x].
                         _params.particle_dia ** 2 * b.bed_voidage ** 3)) +
                        1e2*(
                        1.75*b.gas_phase.properties[t, x].dens_mass *
                        (1 - b.bed_voidage) *
                        (b.velocity_superficial_gas[t, x] +
                         b.velocity_superficial_solid[t]) ** 2 /
                        (b.solid_phase.properties[t, x]._params.particle_dia *
                         b.bed_voidage**3)))
            # The above expression has no absolute values - assumes:
            # (velocity_superficial_gas + velocity_superficial_solid) > 0
        else:
            raise BurntToast(
                    "{} encountered unrecognized argument for "
                    "the pressure drop correlation. Please contact the IDAES"
                    " developers with this bug.".format(self.name))
        # ---------------------------------------------------------------------
        # Reaction contraints

        # Build homogeneous reaction constraints
        if gas_phase.reaction_package is not None:
            # Gas side rate reaction extent
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             gas_phase.reaction_package.rate_reaction_idx,
                             doc="Gas side rate reaction extent")
            def gas_phase_config_rxn_ext(b, t, x, r):
                return 1e3*b.gas_phase.rate_reaction_extent[t, x, r] == 1e3*(
                        b.gas_phase.reactions[t, x].reaction_rate[r] *
                        b.gas_phase.area[t, x])

        # Build hetereogeneous reaction constraints
        if solid_phase.reaction_package is not None:
            # Solid side rate reaction extent
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             solid_phase.reaction_package.rate_reaction_idx,
                             doc="Solid side rate reaction extent")
            def solid_phase_config_rxn_ext(b, t, x, r):
                return 1e3*b.solid_phase.rate_reaction_extent[t, x, r] == 1e3*(
                        b.solid_phase.reactions[t, x].reaction_rate[r] *
                        b.solid_phase.area[t, x])

            # Gas side heterogeneous rate reaction generation
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             gas_phase.property_package.phase_list,
                             gas_phase.property_package.component_list,
                             doc='Gas side heterogeneous'
                             'rate reaction generation')
            def gas_comp_hetero_rxn(b, t, x, p, j):
                return 1e3*b.gas_phase.mass_transfer_term[t, x, p, j] == 1e3*(
                        sum(b.solid_phase.reactions[t, x].
                            rate_reaction_stoichiometry[r, p, j] *
                            b.solid_phase.reactions[t, x].reaction_rate[r]
                            for r in (
                                solid_phase.reaction_package.rate_reaction_idx)
                            ) * b.solid_phase.area[t, x])

        # ---------------------------------------------------------------------
        if self.config.energy_balance_type != EnergyBalanceType.none:
            # Solid phase - gas to solid heat transfer
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Solid phase - gas to solid heat transfer")
            def solid_phase_heat_transfer(b, t, x):
                return (b.solid_phase.heat[t, x] *
                        b.solid_phase.properties[t, x]._params.particle_dia ==
                        6 * b.gas_solid_htc[t, x] *
                        (b.gas_phase.properties[t, x].temperature -
                         b.solid_phase.properties[t, x].temperature) *
                        b.solid_phase.area[t, x])

            # Dimensionless numbers, mass and heat transfer coefficients
            # Particle Reynolds number
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Particle Reynolds number")
            def reynolds_number_particle(b, t, x):
                return (b.Re_particle[t, x] *
                        b.gas_phase.properties[t, x].visc_d ==
                        b.velocity_superficial_gas[t, x] *
                        b.solid_phase.properties[t, x]._params.particle_dia *
                        b.gas_phase.properties[t, x].dens_mass)

            # Prandtl number
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Prandtl number of gas in bed")
            def prandtl_number(b, t, x):
                return (b.Pr[t, x] *
                        b.gas_phase.properties[t, x].therm_cond ==
                        b.solid_phase.properties[t, x].cp_mass *
                        b.gas_phase.properties[t, x].visc_d)

            # Particle Nusselt number
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Particle Nusselt number")
            def nusselt_number_particle(b, t, x):
                return (b.Nu_particle[t, x] ** 3 ==
                        (2.0 + 1.1 * (smooth_abs(b.Re_particle[t, x], b.eps) **
                                      0.6) ** 3) *
                        b.Pr[t, x])

            # Gas-solid heat transfer coefficient
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Gas-solid heat transfer coefficient")
            def gas_solid_htc_eqn(b, t, x):
                return (1e-3*b.gas_solid_htc[t, x] *
                        b.solid_phase.properties[t, x]._params.particle_dia ==
                        1e-3 * b.Nu_particle[t, x] *
                        b.gas_phase.properties[t, x].therm_cond)

            # Gas phase - gas to solid heat transfer
            @self.Constraint(self.flowsheet().time,
                             self.length_domain,
                             doc="Gas phase - gas to solid heat transfer")
            def gas_phase_heat_transfer(b, t, x):
                return (b.gas_phase.heat[t, x] *
                        b.solid_phase.properties[t, x]._params.particle_dia ==
                        -6 * b.gas_solid_htc[t, x] *
                        (b.gas_phase.properties[t, x].temperature -
                         b.solid_phase.properties[t, x].temperature) *
                        b.solid_phase.area[t, x])

        elif self.config.energy_balance_type == EnergyBalanceType.none:
            # If energy balance is none fix gas and solid temperatures to inlet
            @self.Constraint(
                    self.flowsheet().time,
                    self.length_domain,
                    doc="Isothermal gas phase constraint")
            def isothermal_gas_phase(b, t, x):
                if x == self.length_domain.first():
                    return Constraint.Skip
                else:
                    return (
                            b.gas_phase.properties[t, x].temperature ==
                            b.gas_inlet.temperature[t])

            @self.Constraint(
                    self.flowsheet().time,
                    self.length_domain,
                    doc="Isothermal solid phase constraint")
            def isothermal_solid_phase(b, t, x):
                if x == self.length_domain.last():
                    return Constraint.Skip
                else:
                    return (
                            b.solid_phase.properties[t, x].temperature ==
                            b.solid_inlet.temperature[t])

    # =========================================================================
    # Model initialization routine

    def initialize(blk, gas_phase_state_args=None, solid_phase_state_args=None,
                   outlvl=idaeslog.NOTSET, solver=None, optarg=None):
        """
        Initialisation routine for MB unit.

        Keyword Arguments:
            gas_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            solid_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """

        # Set up logger for initialization and solve
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # local aliases used to shorten object names
        gas_phase = blk.config.gas_phase_config
        solid_phase = blk.config.solid_phase_config

        # Keep all unit model geometry constraints, derivative_var constraints,
        # and property block constraints active. Additionaly, in control
        # volumes - keep conservation linking constraints and
        # holdup calculation (for dynamic flowsheets) constraints active

        geometry_constraints_terms = ["bed_area_eqn",
                                      "solid_phase_area",
                                      "gas_phase_area",
                                      "gas_phase_length",
                                      "solid_phase_length"]
        endswith_terms = ("_disc_eq", "linking_constraint",
                          "linking_constraints", "_holdup_calculation")
        startswith_terms = ("properties")

        for c in blk.component_objects(Constraint, descend_into=True):
            if not c.parent_block().local_name.startswith(startswith_terms) \
                and not c.local_name.endswith(endswith_terms) \
                    and c.local_name not in geometry_constraints_terms:
                c.deactivate()

        # ---------------------------------------------------------------------
        # Initialize thermophysical property constraints
        init_log.info('Initialize Thermophysical Properties')
        # Initialize gas_phase block
        gas_phase_flags = blk.gas_phase.properties.initialize(
                                state_args=gas_phase_state_args,
                                outlvl=outlvl,
                                optarg=optarg,
                                solver=solver,
                                hold_state=True)

        # Initialize solid_phase properties block
        solid_phase_flags = blk.solid_phase.properties.initialize(
                                state_args=solid_phase_state_args,
                                outlvl=outlvl,
                                optarg=optarg,
                                solver=solver,
                                hold_state=True)

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Initialize hydrodynamics (gas velocity)
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                calculate_variable_from_constraint(
                    blk.velocity_superficial_gas[t, x],
                    blk.gas_super_vel[t, x])

        blk.gas_super_vel.activate()

        init_log.info('Initialize Hydrodynamics')
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee)
        if results.solver.termination_condition \
                == TerminationCondition.optimal:
            init_log.info_high(
                "Initialization Step 2 {}.".format(
                        idaeslog.condition(results))
                        )
        else:
            _log.warning('{} Initialisation Step 2 Failed.'
                         .format(blk.name))

        # ---------------------------------------------------------------------
        # Initialize mass balance - no reaction and no pressure drop

        # Unfix material balance state variables (including particle porosity)
        # but keep other states fixed
        blk.gas_phase.properties.release_state(
                gas_phase_flags)
        blk.solid_phase.properties.release_state(
                solid_phase_flags)
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                blk.gas_phase.properties[t, x].pressure.fix()
                blk.gas_phase.properties[t, x].temperature.fix()
                blk.solid_phase.properties[t, x].temperature.fix()

        blk.gas_phase.material_balances.activate()

        if gas_phase.reaction_package is not None:
            for t in blk.flowsheet().time:
                gas_rxn_gen = blk.gas_phase.rate_reaction_generation
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_rxn_gen[t, x, p, j].fix(0.0))

        blk.solid_phase.material_balances.activate()
        blk.solid_super_vel.activate()

        if solid_phase.reaction_package is not None:
            for t in blk.flowsheet().time:
                solid_rxn_gen = blk.solid_phase.rate_reaction_generation
                for x in blk.length_domain:
                    for p in solid_phase.property_package.phase_list:
                        for j in solid_phase.property_package.component_list:
                            (solid_rxn_gen[t, x, p, j].fix(0.0))

                # Gas side heterogeneous rate reaction generation
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (blk.gas_phase.mass_transfer_term[t, x, p, j].fix(
                                0.0))

        init_log.info('Initialize Mass Balances')
        init_log.info_high('initialize mass balances - no reactions '
                           'and no pressure drop')
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee)
        if results.solver.termination_condition \
                == TerminationCondition.optimal:
            init_log.info_high(
                "Initialization Step 3a {}.".format(
                        idaeslog.condition(results))
                        )
        else:
            _log.warning('{} Initialisation Step 3a Failed.'
                         .format(blk.name))

        # Initialize mass balance - with reaction and no pressure drop
        if gas_phase.reaction_package is not None:
            # local aliases used to shorten object names
            gas_rxn_gen = blk.gas_phase.rate_reaction_generation
            gas_phase_stoichiometry_eqn = (
                blk.gas_phase.rate_reaction_stoichiometry_constraint)

            # Initialize reaction property package
            blk.gas_phase.reactions.activate()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    obj = blk.gas_phase.reactions[t, x]
                    for c in obj.component_objects(
                            Constraint, descend_into=False):
                        c.activate()

            blk.gas_phase.reactions.initialize(outlvl=outlvl,
                                               optarg=optarg,
                                               solver=solver)

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    for r in gas_phase.reaction_package.rate_reaction_idx:
                        calculate_variable_from_constraint(
                                blk.gas_phase.rate_reaction_extent[t, x, r],
                                blk.gas_phase_config_rxn_ext[t, x, r])
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_rxn_gen[t, x, p, j].unfix())
                            if not (
                                (blk.gas_phase.config.transformation_scheme 
                                    != "FORWARD"
                                 and x == blk.length_domain.first()) or
                                (blk.gas_phase.config.transformation_scheme
                                    == "FORWARD"
                                 and x == blk.length_domain.last())):
                                calculate_variable_from_constraint(
                                    gas_rxn_gen[t, x, p, j],
                                    gas_phase_stoichiometry_eqn[t, x, p, j])

            gas_phase_stoichiometry_eqn.activate()
            blk.gas_phase_config_rxn_ext.activate()

        if solid_phase.reaction_package is not None:
            # local aliases used to shorten object names
            solid_rxn_gen = blk.solid_phase.rate_reaction_generation
            solid_phase_stoichiometry_eqn = (
                blk.solid_phase.rate_reaction_stoichiometry_constraint)
            gas_mass_transfer_term = blk.gas_phase.mass_transfer_term

            # Initialize reaction property package
            blk.solid_phase.reactions.activate()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    obj = blk.solid_phase.reactions[t, x]
                    for c in obj.component_objects(
                            Constraint, descend_into=False):
                        c.activate()

            blk.solid_phase.reactions.initialize(outlvl=outlvl,
                                                 optarg=optarg,
                                                 solver=solver)

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_mass_transfer_term[t, x, p, j].unfix())
                            calculate_variable_from_constraint(
                                gas_mass_transfer_term[t, x, p, j],
                                blk.gas_comp_hetero_rxn[t, x, p, j])
                for x in blk.length_domain:
                    for r in solid_phase.reaction_package.rate_reaction_idx:
                        calculate_variable_from_constraint(
                            blk.solid_phase.rate_reaction_extent[t, x, r],
                            blk.solid_phase_config_rxn_ext[t, x, r])
                    for p in solid_phase.property_package.phase_list:
                        for j in solid_phase.property_package.component_list:
                            (solid_rxn_gen[t, x, p, j].unfix())
                            if not (
                                (blk.solid_phase.config.transformation_scheme 
                                    != "FORWARD"
                                 and x == blk.length_domain.first()) or
                                (blk.solid_phase.config.transformation_scheme
                                    == "FORWARD"
                                 and x == blk.length_domain.last())):
                                calculate_variable_from_constraint(
                                    solid_rxn_gen[t, x, p, j],
                                    solid_phase_stoichiometry_eqn[t, x, p, j])

            blk.gas_comp_hetero_rxn.activate()
            blk.solid_phase.rate_reaction_stoichiometry_constraint.activate()
            blk.solid_phase_config_rxn_ext.activate()

        if (gas_phase.reaction_package is not None or
                solid_phase.reaction_package is not None):
            init_log.info_high('initialize mass balances - with reactions '
                               'and no pressure drop')
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                init_log.info_high(
                    "Initialization Step 3b {}.".format(
                            idaeslog.condition(results))
                            )
            else:
                _log.warning('{} Initialisation Step 3b Failed.'
                             .format(blk.name))

        # Initialize mass balance - with pressure drop
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                # Unfix all pressure variables except at the inlet
                if (blk.gas_phase.properties[t, x].config.defined_state
                        is False):
                    blk.gas_phase.properties[t, x].pressure.unfix()

        blk.gas_phase.pressure_balance.activate()

        # Set scaling factors for pressure balance equation
        blk.gas_phase.scaling_factor_pressure = 1e2

        if blk.config.has_pressure_change:
            blk.gas_phase_config_pressure_drop.activate()

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    calculate_variable_from_constraint(
                        blk.gas_phase.deltaP[t, x],
                        blk.gas_phase_config_pressure_drop[t, x])

            init_log.info_high('initialize mass balances - with reactions '
                               'and pressure drop')
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                init_log.info_high(
                    "Initialization Step 3c {}.".format(
                            idaeslog.condition(results))
                            )
            else:
                _log.warning('{} Initialisation Step 3c Failed.'
                             .format(blk.name))
        # ---------------------------------------------------------------------
        # Initialize energy balance
        if blk.config.energy_balance_type != EnergyBalanceType.none:
            # Initialize dimensionless numbers,
            # mass and heat transfer coefficients
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    calculate_variable_from_constraint(
                        blk.Re_particle[t, x],
                        blk.reynolds_number_particle[t, x])
                    calculate_variable_from_constraint(
                        blk.Pr[t, x],
                        blk.prandtl_number[t, x])
                    calculate_variable_from_constraint(
                        blk.Nu_particle[t, x],
                        blk.nusselt_number_particle[t, x])
                    calculate_variable_from_constraint(
                            blk.gas_solid_htc[t, x],
                            blk.gas_solid_htc_eqn[t, x])
                    calculate_variable_from_constraint(
                        blk.gas_phase.heat[t, x],
                        blk.gas_phase_heat_transfer[t, x])
                    calculate_variable_from_constraint(
                        blk.solid_phase.heat[t, x],
                        blk.solid_phase_heat_transfer[t, x])

            # Unfix temperatures
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    # Unfix all gas temperature variables except at the inlet
                    if (blk.gas_phase.properties[t, x].config.defined_state
                            is False):
                        blk.gas_phase.properties[t, x].temperature.unfix()
                for x in blk.length_domain:
                    # Unfix all solid temperature variables except at the inlet
                    if (blk.solid_phase.properties[t, x].config.defined_state
                            is False):
                        blk.solid_phase.properties[t, x].temperature.unfix()

            blk.reynolds_number_particle.activate()
            blk.prandtl_number.activate()
            blk.nusselt_number_particle.activate()
            blk.gas_solid_htc_eqn.activate()

            # Activate gas phase energy balance equations
            blk.gas_phase_heat_transfer.activate()
            blk.gas_phase.enthalpy_balances.activate()

            # Activate solid phase energy balance equations
            blk.solid_phase_heat_transfer.activate()
            blk.solid_phase.enthalpy_balances.activate()

            init_log.info('Initialize Energy Balances')
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                init_log.info_high(
                    "Initialization Step 4 {}.".format(
                            idaeslog.condition(results))
                            )
            else:
                _log.warning('{} Initialisation Step 4 Failed.'
                             .format(blk.name))

        # Initialize energy balance
        if blk.config.energy_balance_type == EnergyBalanceType.none:
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    # Unfix all gas temperature variables except at the inlet
                    if (blk.gas_phase.properties[t, x].config.defined_state
                            is False):
                        blk.gas_phase.properties[t, x].temperature.unfix()
                for x in blk.length_domain:
                    # Unfix all solid temperature variables except at the inlet
                    if (blk.solid_phase.properties[t, x].config.defined_state
                            is False):
                        blk.solid_phase.properties[t, x].temperature.unfix()

            blk.isothermal_gas_phase.activate()
            blk.isothermal_solid_phase.activate()

            init_log.info('Initialize Energy Balances')
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                init_log.info_high(
                    "Initialization Step 4 {}.".format(
                            idaeslog.condition(results))
                            )
            else:
                _log.warning('{} Initialisation Step 4 Failed.'
                             .format(blk.name))

    def results_plot(blk):
        '''
        Plot method for common moving bed variables

        Variables plotted:
            Tg : Temperature in gas phase
            Ts : Temperature in solid phase
            vg : Superficial gas velocity
            P : Pressure in gas phase
            Ftotal : Total molar flowrate of gas
            Mtotal : Total mass flowrate of solid
            Cg : Concentration of gas components in the gas phase
            y_frac : Mole fraction of gas components in the gas phase
            x_frac : Mass fraction of solid components in the solid phase
        '''
        print()
        print('================================= Reactor plots ==============='
              '==================')
        # local aliases used to shorten object names
        gas_phase = blk.config.gas_phase_config
        solid_phase = blk.config.solid_phase_config

        Tg = []
        Ts = []
        P = []
        Ftotal = []
        Mtotal = []
        vg = []

        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                vg.append(value(blk.velocity_superficial_gas[t, x]))

        fig_vg = plt.figure(1)
        plt.plot(blk.gas_phase.length_domain, vg, label='vg')
        plt.legend(loc=0, ncol=2)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Superficial gas velocity [m/s]")
        fig_vg.savefig('superficial_vel.png')

        # Pressure profile
        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                P.append(blk.gas_phase.properties[t, x].pressure.value)

        fig_P = plt.figure(2)
        plt.plot(blk.gas_phase.length_domain, P, label='P')
        plt.legend(loc=0, ncol=2)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Total Pressure [bar]")
        fig_P.savefig('Pressure.png')

        # Temperature profile
        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                Tg.append(blk.gas_phase.properties[t, x].temperature.value)
            for x in blk.solid_phase.length_domain:
                Ts.append(blk.solid_phase.properties[t, x].temperature.value)
        fig_T = plt.figure(3)
        plt.plot(blk.gas_phase.length_domain, Tg, label='Tg')
        plt.plot(blk.solid_phase.length_domain, Ts, label='Ts')
        plt.legend(loc=0, ncol=2)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Temperature [K]")
        fig_T.savefig('Temperature.png')

        # Bulk gas phase total molar flow rate
        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                Ftotal.append(blk.gas_phase.properties[t, x].flow_mol.value)
        fig_Ftotal = plt.figure(4)
        plt.plot(blk.gas_phase.length_domain, Ftotal)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Total molar gas flow rate [mol/s]")
        fig_Ftotal.savefig('Total_gas_flow.png')

        # Bulk solid phase total mass flow rate
        for t in blk.flowsheet().time:
            for x in blk.solid_phase.length_domain:
                Mtotal.append(blk.solid_phase.properties[t, x].flow_mass.value)
        fig_Mtotal = plt.figure(5)
        plt.plot(blk.solid_phase.length_domain, Mtotal)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Solid total mass flow rate [kg/s]")
        fig_Mtotal.savefig('Total_solid_flow.png')

        # Gas phase mole fractions
        for t in blk.flowsheet().time:
            for j in gas_phase.property_package.component_list:
                y_frac = []
                for x in blk.gas_phase.length_domain:
                    y_frac.append(value(
                        blk.gas_phase.properties[t, x].mole_frac_comp[j]))
                fig_y = plt.figure(6)
                plt.plot(blk.gas_phase.length_domain, y_frac, label=j)
        plt.legend(loc=0, ncol=len(gas_phase.property_package.component_list))
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("y_frac [-]")
        fig_y.savefig('Gas_mole_fractions.png')

        # Solid phase mass fractions
        for t in blk.flowsheet().time:
            for j in solid_phase.property_package.component_list:
                x_frac = []
                for x in blk.solid_phase.length_domain:
                    x_frac.append(value(
                        blk.solid_phase.properties[t, x].mass_frac_comp[j]))
                fig_x = plt.figure(7)
                plt.plot(blk.solid_phase.length_domain, x_frac, label=j)
        plt.legend(loc=0, ncol=len(
            solid_phase.property_package.component_list))
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("x_frac [-]")
        fig_x.savefig('Solid_mass_fractions.png')

        # Gas phase concentrations
        for t in blk.flowsheet().time:
            for j in gas_phase.property_package.component_list:
                Cg = []
                for x in blk.gas_phase.length_domain:
                    Cg.append(
                        blk.gas_phase.properties[t, x].
                        dens_mol_comp[j].value)
                fig_Cg = plt.figure(8)
                plt.plot(blk.gas_phase.length_domain, Cg, label=j)
        plt.legend(loc=0, ncol=len(gas_phase.property_package.component_list))
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Concentration [mol/m3]")
        fig_Cg.savefig('Gas_concentration.png')

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Gas Inlet": self.gas_inlet,
                "Gas Outlet": self.gas_outlet,
                "Solid Inlet": self.solid_inlet,
                "Solid Outlet": self.solid_outlet,
            },
            time_point=time_point,
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
IDAES PLATE HEAT EXCHANGER MODEL (PHE) USING EFFECTIVENESS-NTU METHOD

Notes:
Let P = number of passes(same for both fluids)
if P is even PHE is in counter-current mode
if P is odd PHE  is in co-current mode
Hot and cold fluids flow alternatively in the channels.
Divider plates may be used to partition the passes in separate sections.
The heat transfer area  of a single plate depends on the total heat transfer area
as specified by the manufacturer and the total number of active plates.

Detailed model equations can be found in the paper :
Akula, P., Eslick, J., Bhattacharyya, D. and Miller, D.C., 2019.
Modelling and Parameter Estimation of a Plate Heat Exchanger
as Part of a Solvent-Based Post-Combustion CO2 Capture System.
In Computer Aided Chemical Engineering (Vol. 47, pp. 47-52). Elsevier.

"""

# Import Pyomo libraries
from pyomo.environ import Param, RangeSet, Constraint, Expression,\
    SolverFactory, value, Var, exp, units as pyunits
from pyomo.common.config import ConfigBlock, ConfigValue

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util import get_solver
import idaes.logger as idaeslog

__author__ = "Paul Akula"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("PHE")
class PHEData(UnitModelBlockData):
    """Plate Heat Exchanger(PHE) Unit Model."""

    CONFIG = UnitModelBlockData.CONFIG()

    # Configuration template for fluid specific  arguments
    _SideCONFIG = ConfigBlock()

    CONFIG.declare("passes", ConfigValue(
        default=4,
        domain=int,
        description="Number of passes",
        doc="""Number of passes of the fluids through the heat exchanger"""))

    CONFIG.declare("channel_list", ConfigValue(
        default=[12, 12, 12, 12],
        domain=list,
        description="Number of channels for each pass",
        doc="""Number of channels to be used in each pass where a channel
               is the space between two plates with a flowing fluid"""))

    CONFIG.declare("divider_plate_number", ConfigValue(
        default=0,
        domain=int,
        description="Number of divider plates in heat exchanger",
        doc="""Divider plates are used to create separate partitions in the unit.
               Each pass can be separated by a divider plate"""))

    CONFIG.declare("port_diameter", ConfigValue(
        default=0.2045,
        domain=float,
        description="Diameter of the ports on the plate [m]",
        doc="""Diameter of the ports on the plate for fluid entry/exit
               into a channel"""))

    CONFIG.declare("plate_thermal_cond", ConfigValue(
        default=16.2,
        domain=float,
        description="Thermal conductivity [W/m.K]",
        doc="""Thermal conductivity of the plate material [W/m.K]"""))

    CONFIG.declare("total_area", ConfigValue(
        default=114.3,
        domain=float,
        description="Total heat transfer area [m2]",
        doc="""Total heat transfer area as specifed by the manufacturer"""))

    CONFIG.declare("plate_thickness", ConfigValue(
        default=0.0006,
        domain=float,
        description="Plate thickness [m]",
        doc="""Plate thickness"""))

    CONFIG.declare("plate_vertical_dist", ConfigValue(
        default=1.897,
        domain=float,
        description="Vertical distance between centers of ports [m].",
        doc="""Vertical distance between centers of ports.(Top and bottom ports)
            (approximately equals to the plate length)"""))

    CONFIG.declare("plate_horizontal_dist", ConfigValue(
        default=0.409,
        domain=float,
        description="Horizontal distance between centers of ports [m].",
        doc="""Horizontal distance between centers of ports(Left and right ports)"""))

    CONFIG.declare("plate_pact_length", ConfigValue(
        default=0.381,
        domain=float,
        description="Compressed plate pact length [m].",
        doc="""Compressed plate pact length.
               Length between the Head and the Follower"""))

    CONFIG.declare("surface_enlargement_factor", ConfigValue(
        default=None,
        domain=float,
        description="Surface enlargement factor",
        doc="""Surface enlargement factor is the ratio of single plate area
               (obtained from the total area) to the projected plate area"""))

    CONFIG.declare("plate_gap", ConfigValue(
        default=None,
        domain=float,
        description="Mean channel spacing or gap bewteen two plates [m]",
        doc="""The plate gap is the distance between two adjacent plates that
               forms a flow channel """))

    _SideCONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use ",
        doc="""Property parameter object used to define property calculations
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))

    _SideCONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property package",
        doc="""A ConfigBlock with arguments to be passed to
        property block(s) and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}"""))

    # Create individual config blocks for hot and cold sides
    CONFIG.declare("hot_side",
                   _SideCONFIG(doc="Hot fluid config arguments"))
    CONFIG.declare("cold_side",
                   _SideCONFIG(doc="Cold fluid config arguments"))

    def build(self):
        # Call UnitModel.build to setup model
        super(PHEData, self).build()

        # Consistency check for number of passes and channels in each pass
        for i in self.config.channel_list:
            if not isinstance(i, int):
                raise ConfigurationError("number of channels ({}) must be"
                                         " an integer".format(i))

        if (self.config.passes != len(self.config.channel_list)):
            raise ConfigurationError(
                "The number of elements in the channel list: {}  "
                " does not match the number of passes ({}) given. "
                "Please provide as integers, the number of channels of each pass".format(
                    self.config.channel_list, self.config.passes))

        # ======================================================================
        # Build hot-side  Control Volume (Lean Solvent)
        self.hot_side = ControlVolume0DBlock(default={
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "property_package": self.config.hot_side.property_package,
            "property_package_args": self.config.hot_side.property_package_args})

        self.hot_side.add_state_blocks(has_phase_equilibrium=False)

        self.hot_side.add_material_balances(
            balance_type=MaterialBalanceType.componentTotal,
            has_mass_transfer=False,
            has_phase_equilibrium=False,
            has_rate_reactions=False)

        self.hot_side.add_momentum_balances(
            balance_type=MomentumBalanceType.pressureTotal,
            has_pressure_change=True)

        # Energy balance is based on the effectiveness Number of Transfer units
        # (E-NTU method) and inluded as performance equations. Hence the control
        #  volume energy balances are not added.

        # ======================================================================
        # Build cold-side  Control Volume(Rich solvent)
        self.cold_side = ControlVolume0DBlock(default={
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "property_package": self.config.cold_side.property_package,
            "property_package_args": self.config.cold_side.property_package_args})

        self.cold_side.add_state_blocks(has_phase_equilibrium=False)

        self.cold_side.add_material_balances(
            balance_type=MaterialBalanceType.componentTotal,
            has_mass_transfer=False,
            has_phase_equilibrium=False,
            has_rate_reactions=False)

        self.cold_side.add_momentum_balances(
            balance_type=MomentumBalanceType.pressureTotal,
            has_pressure_change=True)

        # ======================================================================
        # Add Ports to control volumes
        # hot-side
        self.add_inlet_port(name="hot_inlet",
                            block=self.hot_side, doc='inlet Port')
        self.add_outlet_port(name="hot_outlet",
                             block=self.hot_side, doc='outlet Port')

        # cold-side
        self.add_inlet_port(name="cold_inlet",
                            block=self.cold_side, doc='inlet Port')
        self.add_outlet_port(name="cold_outlet",
                             block=self.cold_side, doc='outlet Port')
        # ======================================================================
        # Add performace equation method
        self._make_params()
        self._make_performance_method()

    def _make_params(self):
        self.P = Param(initialize=self.config.passes,
                       units=None,
                       doc="Total number of passes for hot or cold fluid")
        self.PH = RangeSet(self.P,
                           doc="Set of hot fluid passes")
        self.PC = RangeSet(self.P,
                           doc="Set of cold fluid passes(equal to PH)")

        self.plate_thermal_cond = Param(mutable=True,
                                        initialize=self.config.plate_thermal_cond,
                                        units=pyunits.W / pyunits.m / pyunits.K,
                                        doc="Plate thermal conductivity")
        self.plate_thick = Param(mutable=True,
                                 initialize=self.config.plate_thickness,
                                 units=pyunits.m,
                                 doc="Plate thickness")

        self.port_dia = Param(mutable=True,
                              initialize=self.config.port_diameter,
                              units=pyunits.m,
                              doc=" Port diameter of plate ")
        self.Np = Param(self.PH,
                        units=None,
                        doc="Number of channels in each pass", mutable=True)
        # Number of channels in each pass
        for i in self.PH:
            self.Np[i].value = self.config.channel_list[i - 1]

        # ---------------------------------------------------------------------
        # Assign plate specifications

        # effective plate length & width
        _effective_plate_length = self.config.plate_vertical_dist - \
            self.config.port_diameter
        _effective_plate_width = self.config.plate_horizontal_dist + \
            self.config.port_diameter

        self.plate_length = Expression(expr=_effective_plate_length)
        self.plate_width = Expression(expr=_effective_plate_width)

        # Area of single plate
        _total_active_plate_number = 2 * sum(self.config.channel_list) - 1 -\
            self.config.divider_plate_number

        self.plate_area = Expression(expr=self.config.total_area /
                                     _total_active_plate_number,
                                     doc="Heat transfer area of single plate")

        # Plate gap
        if self.config.plate_gap is None:
            _total_plate_number = 2 * sum(self.config.channel_list) + 1 +\
                self.config.divider_plate_number
            _plate_pitch = self.config.plate_pact_length / _total_plate_number

            _plate_gap = _plate_pitch - self.config.plate_thickness
        else:
            _plate_gap = self.config.plate_gap

        self.plate_gap = Expression(expr=_plate_gap)

        # Surface enlargement factor
        if self.config.surface_enlargement_factor is None:
            _projected_plate_area = _effective_plate_length * _effective_plate_width
            _surface_enlargement_factor = self.plate_area / _projected_plate_area
        else:
            _surface_enlargement_factor = self.config.surface_enlargement_factor

        self.surface_enlargement_factor = Expression(expr=_surface_enlargement_factor)

        # Channel equivalent diameter
        self.channel_dia = Expression(expr=2 * self.plate_gap /
                                      _surface_enlargement_factor ,
                                      doc=" Channel equivalent diameter")

        # heat transfer parameters
        self.param_a = Var(initialize=0.3, bounds=(0.2, 0.4), units=None,
                           doc='Nusselt parameter')
        self.param_b = Var(initialize=0.663, bounds=(0.3, 0.7), units=None,
                           doc='Nusselt parameter')
        self.param_c = Var(initialize=1 / 3.0, bounds=(1e-5, 2), units=None,
                           doc='Nusselt parameter')
        self.param_a.fix(0.4)
        self.param_b.fix(0.663)
        self.param_c.fix(0.333)

    def _make_performance_method(self):

        solvent_list = self.config.hot_side.property_package.component_list_solvent

        def rule_trh(blk, t):
            return (blk.hot_side.properties_out[t].temperature /
                    blk.hot_side.properties_in[t].temperature)
        self.trh = Expression(self.flowsheet().time, rule=rule_trh,
                              doc='Ratio of hot outlet temperature to hot'
                                  'inlet temperature')

        def rule_trc(blk, t):
            return (blk.cold_side.properties_out[t].temperature /
                    blk.cold_side.properties_in[t].temperature)
        self.trc = Expression(self.flowsheet().time, rule=rule_trc,
                              doc='Ratio of cold outlet temperature to cold'
                                  ' inlet temperature')

        def rule_cp_comp_hot(blk, t, j):
            return 1e3 * (
                blk.hot_side.properties_in[t]._params.cp_param[j, 1] +
                blk.hot_side.properties_in[t]._params.cp_param[j, 2] / 2 *
                blk.hot_side.properties_in[t].temperature *
                (blk.trh[t] + 1) +
                blk.hot_side.properties_in[t]._params.cp_param[j, 3] / 3 *
                (blk.hot_side.properties_in[t].temperature**2) *
                (blk.trh[t]**2 + blk.trh[t] + 1) +
                blk.hot_side.properties_in[t]._params.cp_param[j, 4] / 4 *
                (blk.hot_side.properties_in[t].temperature**3) *
                (blk.trh[t] + 1) * (blk.trh[t]**2 + 1) +
                blk.hot_side.properties_in[t]._params.cp_param[j, 5] / 5 *
                (blk.hot_side.properties_in[t].temperature**4) *
                (blk.trh[t]**4 + blk.trh[t]**3 + blk.trh[t]**2 + blk.trh[t] + 1))

        self.cp_comp_hot = Expression(self.flowsheet().time,
                                      solvent_list,
                                      rule=rule_cp_comp_hot,
                                      doc='Component mean specific heat capacity'
                                      ' btw inlet and outlet'
                                      ' of hot-side temperature')

        def rule_cp_hot(blk, t):
            return sum(blk.cp_comp_hot[t, j] *
                       blk.hot_side.properties_in[t].mass_frac_co2_free[j]
                       for j in solvent_list)
        self.cp_hot = Expression(self.flowsheet().time, rule=rule_cp_hot,
                                 doc='Hot-side mean specific heat capacity on'
                                     'free CO2 basis')

        def rule_cp_comp_cold(blk, t, j):
            return 1e3 * (
                blk.cold_side.properties_in[t]._params.cp_param[j, 1] +
                blk.cold_side.properties_in[t]._params.cp_param[j, 2] / 2 *
                blk.cold_side.properties_in[t].temperature *
                (blk.trc[t] + 1) +
                blk.cold_side.properties_in[t]._params.cp_param[j, 3] / 3 *
                (blk.cold_side.properties_in[t].temperature**2) *
                (blk.trc[t]**2 + blk.trc[t] + 1) +
                blk.cold_side.properties_in[t]._params.cp_param[j, 4] / 4 *
                (blk.cold_side.properties_in[t].temperature**3) *
                (blk.trc[t] + 1) * (blk.trc[t]**2 + 1) +
                blk.cold_side.properties_in[t]._params.cp_param[j, 5] / 5 *
                (blk.cold_side.properties_in[t].temperature**4) *
                (blk.trc[t]**4 + blk.trc[t]**3 + blk.trc[t]**2 + blk.trc[t] + 1))

        self.cp_comp_cold = Expression(self.flowsheet().time,
                                       solvent_list,
                                       rule=rule_cp_comp_cold,
                                       doc='Component mean specific heat capacity'
                                           'btw inlet and outlet'
                                           ' of cold-side temperature')

        def rule_cp_cold(blk, t):
            return sum(blk.cp_comp_cold[t, j] *
                       blk.cold_side.properties_in[t].mass_frac_co2_free[j]
                       for j in solvent_list)
        self.cp_cold = Expression(self.flowsheet().time, rule=rule_cp_cold,
                                  doc='Cold-side mean specific heat capacity'
                                      'on free CO2 basis')

        # Model Variables
        self.Th_in = Var(self.flowsheet().time,
                         self.PH, initialize=393, units=pyunits.K,
                         doc="Hot Temperature IN of pass")
        self.Th_out = Var(self.flowsheet().time,
                          self.PH, initialize=325, units=pyunits.K,
                          doc="Hot Temperature OUT of pass")
        self.Tc_in = Var(self.flowsheet().time,
                         self.PH, initialize=320, units=pyunits.K,
                         doc="Cold Temperature IN of pass")
        self.Tc_out = Var(self.flowsheet().time,
                          self.PH, initialize=390, units=pyunits.K,
                          doc="Cold Temperature OUT of pass")

        # ======================================================================
        # PERFORMANCE EQUATIONS
        # mass flow rate in kg/s
        def rule_mh_in(blk, t):
            return blk.hot_side.properties_in[t].flow_mol *\
                blk.hot_side.properties_in[t].mw
        self.mh_in = Expression(self.flowsheet().time, rule=rule_mh_in,
                                doc='Hotside mass flow rate [kg/s]')

        def rule_mc_in(blk, t):
            return blk.cold_side.properties_in[t].flow_mol *\
                blk.cold_side.properties_in[t].mw
        self.mc_in = Expression(self.flowsheet().time, rule=rule_mc_in,
                                doc='Coldside mass flow rate [kg/s]')

        # ----------------------------------------------------------------------
        # port mass velocity[kg/m2.s]
        def rule_Gph(blk, t):
            return (4 * blk.mh_in[t] * 7) / (22 * blk.port_dia**2)

        self.Gph = Expression(self.flowsheet().time, rule=rule_Gph,
                              doc='Hotside port mass velocity[kg/m2.s]')

        def rule_Gpc(blk, t):
            return (4 * blk.mc_in[t] * 7) / (22 * blk.port_dia**2)

        self.Gpc = Expression(self.flowsheet().time, rule=rule_Gpc,
                              doc='Coldside port mass velocity[kg/m2.s]')

        # ----------------------------------------------------------------------
        # Reynold & Prandtl numbers
        def rule_Re_h(blk, t, p):
            return blk.mh_in[t] * blk.channel_dia /\
                (blk.Np[p] * blk.plate_width *
                 blk.plate_gap * blk.hot_side.properties_in[t].visc_d)
        self.Re_h = Expression(self.flowsheet().time, self.PH,
                               rule=rule_Re_h,
                               doc='Hotside Reynolds number')

        def rule_Re_c(blk, t, p):
            return blk.mc_in[t] * blk.channel_dia /\
                (blk.Np[p] * blk.plate_width *
                 blk.plate_gap * blk.cold_side.properties_in[t].visc_d)
        self.Re_c = Expression(self.flowsheet().time, self.PH,
                               rule=rule_Re_c,
                               doc='Coldside Reynolds number')

        def rule_Pr_h(blk, t):
            return blk.cp_hot[t] * blk.hot_side.properties_in[t].visc_d /\
                blk.hot_side.properties_in[t].thermal_cond
        self.Pr_h = Expression(self.flowsheet().time,
                               rule=rule_Pr_h,
                               doc='Hotside Prandtl number')

        def rule_Pr_c(blk, t):
            return blk.cp_cold[t] * blk.cold_side.properties_in[t].visc_d /\
                blk.cold_side.properties_in[t].thermal_cond
        self.Pr_c = Expression(self.flowsheet().time,
                               rule=rule_Pr_c,
                               doc='Coldside Prandtl number')
        # ----------------------------------------------------------------------
        # Film heat transfer coefficients

        def rule_hotside_transfer_coef(blk, t, p):
            return (blk.hot_side.properties_in[t].thermal_cond / blk.channel_dia *
                    blk.param_a * blk.Re_h[t, p]**blk.param_b *
                    blk.Pr_h[t]**blk.param_c)

        self.h_hot = Expression(self.flowsheet().time,
                                self.PH, rule=rule_hotside_transfer_coef,
                                doc='Hotside heat transfer coefficient')

        def rule_coldside_transfer_coef(blk, t, p):
            return (blk.cold_side.properties_in[t].thermal_cond / blk.channel_dia *
                    blk.param_a * blk.Re_c[t, p]**blk.param_b *
                    blk.Pr_c[t]**blk.param_c)

        self.h_cold = Expression(self.flowsheet().time,
                                 self.PH, rule=rule_coldside_transfer_coef,
                                 doc='Coldside heat transfer coefficient')

        # ----------------------------------------------------------------------
        # Friction factor calculation
        def rule_fric_h(blk, t):
            return 18.29 * blk.Re_h[t, 1]**(-0.652)
        self.fric_h = Expression(self.flowsheet().time,
                                 rule=rule_fric_h,
                                 doc='Hotside friction factor')

        def rule_fric_c(blk, t):
            return 1.441 * self.Re_c[t, 1]**(-0.206)
        self.fric_c = Expression(self.flowsheet().time,
                                 rule=rule_fric_c,
                                 doc='Coldside friction factor')

        # ----------------------------------------------------------------------
        # pressure drop calculation
        def rule_hotside_dP(blk, t):
            return (2 * blk.fric_h[t] * (blk.plate_length + blk.port_dia) *
                    blk.P * blk.Gph[t]**2) /\
                (blk.hot_side.properties_in[t].dens_mass *
                 blk.channel_dia) + 1.4 * blk.P * blk.Gph[t]**2 * 0.5 /\
                blk.hot_side.properties_in[t].dens_mass + \
                blk.hot_side.properties_in[t].dens_mass * \
                9.81 * (blk.plate_length + blk.port_dia)

        self.dP_h = Expression(self.flowsheet().time,
                               rule=rule_hotside_dP,
                               doc='Hotside pressure drop  [Pa]')

        def rule_coldside_dP(blk, t):
            return (2 * blk.fric_c[t] * (blk.plate_length + blk.port_dia) *
                    blk.P * blk.Gpc[t]**2) /\
                (blk.cold_side.properties_in[t].dens_mass * blk.channel_dia) +\
                1.4 * (blk.P * blk.Gpc[t]**2 * 0.5 /
                       blk.cold_side.properties_in[t].dens_mass) + \
                blk.cold_side.properties_in[t].dens_mass * \
                9.81 * (blk.plate_length + blk.port_dia)

        self.dP_c = Expression(self.flowsheet().time,
                               rule=rule_coldside_dP,
                               doc='Coldside pressure drop  [Pa]')

        def rule_eq_deltaP_hot(blk, t):
            return blk.hot_side.deltaP[t] == -blk.dP_h[t]
        self.eq_deltaP_hot = Constraint(self.flowsheet().time,
                                        rule=rule_eq_deltaP_hot)

        def rule_eq_deltaP_cold(blk, t):
            return blk.cold_side.deltaP[t] == -blk.dP_c[t]
        self.eq_deltaP_cold = Constraint(self.flowsheet().time,
                                         rule=rule_eq_deltaP_cold)

        # ----------------------------------------------------------------------
        # Overall heat transfer coefficients
        def rule_U(blk, t, p):
            return 1.0 /\
                (1.0 / blk.h_hot[t, p] + blk.plate_gap / blk.plate_thermal_cond +
                 1.0 / blk.h_cold[t, p])
        self.U = Expression(self.flowsheet().time, self.PH,
                            rule=rule_U,
                            doc='Overall heat transfer coefficient')
        # ----------------------------------------------------------------------
        # capacitance of hot and cold fluid

        def rule_Caph(blk, t, p):
            return blk.mh_in[t] * blk.cp_hot[t] / blk.Np[p]
        self.Caph = Expression(self.flowsheet().time, self.PH,
                               rule=rule_Caph,
                               doc='Hotfluid capacitance rate')

        def rule_Capc(blk, t, p):
            return blk.mc_in[t] * blk.cp_cold[t] / blk.Np[p]
        self.Capc = Expression(self.flowsheet().time, self.PH,
                               rule=rule_Capc,
                               doc='Coldfluid capacitance rate')

        # ----------------------------------------------------------------------
        # min n max capacitance and capacitance ratio
        def rule_Cmin(blk, t, p):
            return 0.5 * (blk.Caph[t, p] + blk.Capc[t, p] -
                          ((blk.Caph[t, p] - blk.Capc[t, p])**2 + 0.00001)**0.5)
        self.Cmin = Expression(self.flowsheet().time, self.PH,
                               rule=rule_Cmin,
                               doc='Minimum capacitance rate')

        def rule_Cmax(blk, t, p):
            return 0.5 * (blk.Caph[t, p] + blk.Capc[t, p] +
                          ((blk.Caph[t, p] - blk.Capc[t, p])**2 + 0.00001)**0.5)
        self.Cmax = Expression(self.flowsheet().time, self.PH,
                               rule=rule_Cmax,
                               doc='Maximum capacitance rate')

        def rule_CR(blk, t, p):
            return blk.Cmin[t, p] / blk.Cmax[t, p]
        self.CR = Expression(self.flowsheet().time, self.PH,
                             rule=rule_CR,
                             doc='Capacitance ratio')

        # ----------------------------------------------------------------------
        # Number of Transfer units for sub heat exchanger
        def rule_NTU(blk, t, p):
            return blk.U[t, p] * blk.plate_area / blk.Cmin[t, p]
        self.NTU = Expression(self.flowsheet().time, self.PH,
                              rule=rule_NTU,
                              doc='Number of Transfer Units')

        # ----------------------------------------------------------------------
        # effectiveness of sub-heat exchangers
        def rule_Ecf(blk, t, p):
            if blk.P.value % 2 == 0:
                return (1 - exp(-blk.NTU[t, p] * (1 - blk.CR[t, p]))) / \
                    (1 - blk.CR[t, p] *
                     exp(-blk.NTU[t, p] * (1 - blk.CR[t, p])))
            elif blk.P.value % 2 == 1:
                return (1 - exp(-blk.NTU[t, p] * (1 + blk.CR[t, p]))) / (1 + blk.CR[t, p])

        self.Ecf = Expression(self.flowsheet().time, self.PH,
                              rule=rule_Ecf,
                              doc='Effectiveness for sub-HX')

        # ----------------------------------------------------------------------
        # Energy balance equations for hot fluid in sub-heat exhanger
        def rule_Ebh_eq(blk, t, p):
            return blk.Th_out[t, p] == blk.Th_in[t, p] -\
                blk.Ecf[t, p] * blk.Cmin[t, p] / blk.Caph[t, p] * \
                (blk.Th_in[t, p] - blk.Tc_in[t, p])

        self.Ebh_eq = Constraint(self.flowsheet().time, self.PH,
                                 rule=rule_Ebh_eq,
                                 doc='Hot fluid sub-heat exchanger energy balance')

        # Hot fluid exit temperature
        def rule_Tout_hot(blk, t):
            return blk.Th_out[t, blk.P.value] ==\
                blk.hot_side.properties_out[t].temperature

        self.Tout_hot_eq = Constraint(self.flowsheet().time,
                                      rule=rule_Tout_hot,
                                      doc='Hot fluid exit temperature')

        # Energy balance equations for cold fluid in sub-heat exhanger
        def rule_Ebc_eq(blk, t, p):
            return blk.Tc_out[t, p] == blk.Tc_in[t, p] + \
                blk.Ecf[t, p] * blk.Cmin[t, p] / blk.Capc[t, p] * \
                (blk.Th_in[t, p] - blk.Tc_in[t, p])

        self.Ebc_eq = Constraint(self.flowsheet().time, self.PH,
                                 rule=rule_Ebc_eq,
                                 doc='Cold fluid sub-heat exchanger energy balance')

        # Cold fluid exit temperature
        def rule_Tout_cold(blk, t):
            return blk.Tc_out[t, 1] ==\
                blk.cold_side.properties_out[t].temperature

        self.Tout_cold_eq = Constraint(self.flowsheet().time,
                                       rule=rule_Tout_cold,
                                       doc='Cold fluid exit temperature')

        # ----------------------------------------------------------------------
        # Energy balance boundary conditions
        def rule_hot_BCIN(blk, t):
            return blk.Th_in[t, 1] == \
                blk.hot_side.properties_in[t].temperature

        self.hot_BCIN = Constraint(self.flowsheet().time,
                                   rule=rule_hot_BCIN,
                                   doc='Hot fluid inlet boundary conditions')

        def rule_cold_BCIN(blk, t):
            return blk.Tc_in[t, blk.P.value] ==\
                blk.cold_side.properties_in[t].temperature
        self.cold_BCIN = Constraint(self.flowsheet().time,
                                    rule=rule_cold_BCIN,
                                    doc='Cold fluid inlet boundary conditions')

        Pset = [i for i in range(1, self.P.value)]

        def rule_hot_BC(blk, t, p):
            return blk.Th_out[t, p] == blk.Th_in[t, p + 1]
        self.hot_BC = Constraint(self.flowsheet().time, Pset,
                                 rule=rule_hot_BC,
                                 doc='Hot fluid boundary conditions: change of pass')

        def rule_cold_BC(blk, t, p):
            return blk.Tc_out[t, p + 1] == blk.Tc_in[t, p]
        self.cold_BC = Constraint(self.flowsheet().time, Pset,
                                  rule=rule_cold_BC,
                                  doc='Cold fluid boundary conditions: change of pass')

        # ----------------------------------------------------------------------
        # Energy transferred
        def rule_QH(blk, t):
            return blk.mh_in[t] * blk.cp_hot[t] *\
                (blk.hot_side.properties_in[t].temperature -
                 blk.hot_side.properties_out[t].temperature)
        self.QH = Expression(self.flowsheet().time, rule=rule_QH,
                             doc='Heat lost by hot fluid')

        def rule_QC(blk, t):
            return blk.mc_in[t] * blk.cp_cold[t] *\
                (blk.cold_side.properties_out[t].temperature -
                 blk.cold_side.properties_in[t].temperature)
        self.QC = Expression(self.flowsheet().time, rule=rule_QH,
                             doc='Heat gain by cold fluid')

    def initialize(blk, hotside_state_args=None, coldside_state_args=None,
                   outlvl=idaeslog.NOTSET, solver=None, optarg=None):
        '''
        Initialisation routine for PHE unit (default solver ipopt)

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None)
        Returns:
            None
        '''
        # Set solver options
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag='unit')
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        hotside_state_args = {
            'flow_mol': value(blk.hot_inlet.flow_mol[0]),
            'temperature': value(blk.hot_inlet.temperature[0]),
            'pressure': value(blk.hot_inlet.pressure[0]),
            'mole_frac_comp':
            {'H2O': value(blk.hot_inlet.mole_frac_comp[0, 'H2O']),
             'CO2': value(blk.hot_inlet.mole_frac_comp[0, 'CO2']),
             'MEA': value(blk.hot_inlet.mole_frac_comp[0, 'MEA'])}}

        coldside_state_args = {
            'flow_mol': value(blk.cold_inlet.flow_mol[0]),
            'temperature': value(blk.cold_inlet.temperature[0]),
            'pressure': value(blk.cold_inlet.pressure[0]),
            'mole_frac_comp':
            {'H2O': value(blk.cold_inlet.mole_frac_comp[0, 'H2O']),
             'CO2': value(blk.cold_inlet.mole_frac_comp[0, 'CO2']),
             'MEA': value(blk.cold_inlet.mole_frac_comp[0, 'MEA'])}}

        # ---------------------------------------------------------------------
        # Initialize the INLET properties
        init_log.info('STEP 1: PROPERTY INITIALIZATION')
        init_log.info_high("INLET Properties initialization")
        blk.hot_side.properties_in.initialize(
            state_args=hotside_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True)
        blk.cold_side.properties_in.initialize(
            state_args=coldside_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True)

        # Initialize the OUTLET properties
        init_log.info_high("OUTLET Properties initialization")
        blk.hot_side.properties_out.initialize(
            state_args=hotside_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=False)

        blk.cold_side.properties_out.initialize(
            state_args=coldside_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=False)
        # ----------------------------------------------------------------------
        init_log.info('STEP 2: PHE INITIALIZATION')
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("STEP 2 Complete: {}.".format(idaeslog.condition(res)))
        init_log.info('INITIALIZATION COMPLETED')



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Python script to read costing components
This script reads the library of costing components (scaled cost, reference
parameters, costing exponents, etc.) from the json files.
First, open json file, then create a python dictionary that gets imported into
power_plant_costing.py

Three python dictionaries that are loaded:
* BB_costing_exponents
* BB_costing_params
* sCO2_costing_params


"""
__author__ = "Costing Team (A. Noring and M. Zamarripa)"
__version__ = "1.0.0"

import os
import json
from pyomo.common.fileutils import this_file_dir

directory = this_file_dir()


''' The costing exponents dictionary contains information from the QGESS on
capital cost scaling methodology (DOE/NETL-2019/1784). Specifically it includes
scaling exponents, valid ranges for the scaled parameter, and units for those
ranges. It is important to note the units only apply to the ranges and are not
neccessarily the units that the reference parameter value will be given in.
This dictionary is nested with the following structure:
    tech type --> account --> property name --> property value'''

with open(os.path.join(directory, "BB_costing_exponents.json"), 'r') as file:
    BB_costing_exponents = json.load(file)


'''
The costing params dictionary contains information from the BBR4 COE
spreadsheet. It includes the total plant cost (TPC), reference parameter value,
and units for that value.

Some accounts are costed using two different reference
parameters, these accounts have been divided into two separate accounts
following
the naming convention x.x.a and x.x.b.
This dictionary is nested with the following structure:
tech type --> CCS --> account --> property name --> property values'''
with open(os.path.join(directory,"BB_costing_parameters.json"), 'r') as file:
    BB_costing_params = json.load(file)


with open(os.path.join(directory, "sCO2_costing_parameters.json"), 'r') as file:
    sCO2_costing_params = json.load(file)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Author: Andrew Lee, Alejandro Garciadiego
"""

import pytest
from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Set,
                           SolverStatus,
                           TerminationCondition,
                           value,
                           Var,
                           units as pyunits)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import (MaterialBalanceType,
                        EnergyBalanceType,
                        MaterialFlowBasis,
                        Component)
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              fixed_variables_set,
                                              activated_constraints_set)
from idaes.core.util import get_solver

from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock)

from idaes.generic_models.properties.core.state_definitions import FTPx

from idaes.generic_models.properties.core.examples.HC_PR_vap \
    import configuration_vap


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# Test for configuration dictionaries with parameters from Properties of Gases
# and liquids 4th edition
class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(default=configuration_vap)

        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 1
        for i in model.params.phase_list:
            assert i in ["Vap"]
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 13
        for i in model.params.component_list:
            assert i in ['hydrogen',
                         'methane',
                         'ethane',
                         'propane',
                         'nbutane',
                         'ibutane',
                         'ethylene',
                         'propene',
                         'butene',
                         'pentene',
                         'hexene',
                         'heptene',
                         'octene']

            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 13
        for i in model.params._phase_component_set:
            assert i in [
                ("Vap", "hydrogen"), ("Vap", "methane"), ("Vap", "ethane"),
                ("Vap", "propane"), ("Vap", "nbutane"), ("Vap", "ibutane"),
                ("Vap", "ethylene"), ("Vap", "propene"), ("Vap", "butene"),
                ("Vap", "pentene"), ("Vap", "hexene"), ("Vap", "heptene"),
                ("Vap", "octene")]

        assert model.params.config.state_definition == FTPx

        assertStructuredAlmostEqual(
            model.params.config.state_bounds,
            { "flow_mol": (0, 100, 1000, pyunits.mol/pyunits.s),
              "temperature": (273.15, 300, 1500, pyunits.K),
              "pressure": (5e4, 1e5, 1e7, pyunits.Pa) },
            item_callback=lambda x: value(x) * (
                pyunits.get_units(x) or pyunits.dimensionless)._get_pint_unit()
        )

        assert model.params.pressure_ref.value == 101325
        assert model.params.temperature_ref.value == 298.15

        assert model.params.hydrogen.mw.value == 2.016E-3
        assert model.params.hydrogen.pressure_crit.value == 12.9e5
        assert model.params.hydrogen.temperature_crit.value == 33.2

        assert model.params.methane.mw.value == 16.043E-3
        assert model.params.methane.pressure_crit.value == 46e5
        assert model.params.methane.temperature_crit.value == 190.4

        assert model.params.ethane.mw.value == 30.070E-3
        assert model.params.ethane.pressure_crit.value == 48.8e5
        assert model.params.ethane.temperature_crit.value == 305.4

        assert model.params.propane.mw.value == 44.094E-3
        assert model.params.propane.pressure_crit.value == 42.5e5
        assert model.params.propane.temperature_crit.value == 369.8

        assert model.params.nbutane.mw.value == 58.124E-3
        assert model.params.nbutane.pressure_crit.value == 38.0e5
        assert model.params.nbutane.temperature_crit.value == 425.2

        assert model.params.ibutane.mw.value == 58.124E-3
        assert model.params.ibutane.pressure_crit.value == 36.5e5
        assert model.params.ibutane.temperature_crit.value == 408.2

        assert model.params.ethylene.mw.value == 28.054E-3
        assert model.params.ethylene.pressure_crit.value == 50.5e5
        assert model.params.ethylene.temperature_crit.value == 282.4

        assert model.params.propene.mw.value == 42.081E-3
        assert model.params.propene.pressure_crit.value == 46.2e5
        assert model.params.propene.temperature_crit.value == 365.0

        assert model.params.butene.mw.value == 56.104E-3
        assert model.params.butene.pressure_crit.value == 40.2e5
        assert model.params.butene.temperature_crit.value == 419.3

        assert model.params.pentene.mw.value == 70.135E-3
        assert model.params.pentene.pressure_crit.value == 40.5e5
        assert model.params.pentene.temperature_crit.value == 464.7

        assert model.params.hexene.mw.value == 84.162E-3
        assert model.params.hexene.pressure_crit.value == 31.7e5
        assert model.params.hexene.temperature_crit.value == 504.0

        assert model.params.heptene.mw.value == 98.189E-3
        assert model.params.heptene.pressure_crit.value == 25.4e5
        assert model.params.heptene.temperature_crit.value == 537.2

        assert model.params.octene.mw.value == 112.216E-3
        assert model.params.octene.pressure_crit.value == 26.2e5
        assert model.params.octene.temperature_crit.value == 566.6
        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(default=configuration_vap)

        model.props = model.params.build_state_block(
                [1],
                default={"defined_state": True})

        # Fix state
        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(295.00)
        model.props[1].pressure.fix(1e5)
        model.props[1].mole_frac_comp["hydrogen"].fix(0.077)
        model.props[1].mole_frac_comp["methane"].fix(0.077)
        model.props[1].mole_frac_comp["ethane"].fix(0.077)
        model.props[1].mole_frac_comp["propane"].fix(0.077)
        model.props[1].mole_frac_comp["nbutane"].fix(0.077)
        model.props[1].mole_frac_comp["ibutane"].fix(0.077)
        model.props[1].mole_frac_comp["ethylene"].fix(0.077)
        model.props[1].mole_frac_comp["propene"].fix(0.077)
        model.props[1].mole_frac_comp["butene"].fix(0.077)
        model.props[1].mole_frac_comp["pentene"].fix(0.077)
        model.props[1].mole_frac_comp["hexene"].fix(0.077)
        model.props[1].mole_frac_comp["heptene"].fix(0.077)
        model.props[1].mole_frac_comp["octene"].fix(0.076)

        assert degrees_of_freedom(model.props[1]) == 0

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.props[1].flow_mol, Var)
        assert value(model.props[1].flow_mol) == 1
        assert model.props[1].flow_mol.ub == 1000
        assert model.props[1].flow_mol.lb == 0

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 1e5
        assert model.props[1].pressure.ub == 1e7
        assert model.props[1].pressure.lb == 5e4

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 295
        assert model.props[1].temperature.ub == 1500
        assert model.props[1].temperature.lb == 273.15

        assert isinstance(model.props[1].mole_frac_comp, Var)
        assert len(model.props[1].mole_frac_comp) == 13
        for i in model.props[1].mole_frac_comp:
            assert value(model.props[1].mole_frac_comp[i]) == \
                pytest.approx(0.077, abs=1e-2)

        # Check supporting variables
        assert isinstance(model.props[1].flow_mol_phase, Var)
        assert len(model.props[1].flow_mol_phase) == 1

        assert isinstance(model.props[1].mole_frac_phase_comp, Var)
        assert len(model.props[1].mole_frac_phase_comp) == 13

        assert isinstance(model.props[1].phase_frac, Var)
        assert len(model.props[1].phase_frac) == 1

        assert isinstance(model.props[1].total_flow_balance, Constraint)
        assert len(model.props[1].total_flow_balance) == 1

        assert isinstance(model.props[1].component_flow_balances, Constraint)
        assert len(model.props[1].component_flow_balances) == 13

        assert isinstance(model.props[1].phase_fraction_constraint, Constraint)
        assert len(model.props[1].phase_fraction_constraint) == 1

        assert_units_consistent(model)

    @pytest.mark.unit
    def test_get_material_flow_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert str(model.props[1].get_material_flow_terms(p, j)) == \
                    str(model.props[1].flow_mol_phase_comp[p, j])

    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, model):
        for p in model.params.phase_list:
            assert str(model.props[1].get_enthalpy_flow_terms(p)) == str(
                model.props[1]._enthalpy_flow_term[p])

    @pytest.mark.unit
    def test_get_material_density_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                if j == "hydrogen":
                    assert str(
                        model.props[1].get_material_density_terms(p, j)) == \
                        str(model.props[1]._material_density_term[p, j])

    @pytest.mark.unit
    def test_get_energy_density_terms(self, model):
        for p in model.params.phase_list:
            assert str(model.props[1].get_energy_density_terms(p)) == str(
                model.props[1]._energy_density_term[p])

    @pytest.mark.unit
    def test_default_material_balance_type(self, model):
        assert model.props[1].default_material_balance_type() == \
            MaterialBalanceType.componentTotal

    @pytest.mark.unit
    def test_default_energy_balance_type(self, model):
        assert model.props[1].default_energy_balance_type() == \
            EnergyBalanceType.enthalpyTotal

    @pytest.mark.unit
    def test_get_material_flow_basis(self, model):
        assert model.props[1].get_material_flow_basis() == \
            MaterialFlowBasis.molar

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol",
                         "mole_frac_comp",
                         "temperature",
                         "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol",
                         "mole_frac_comp",
                         "temperature",
                         "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["Total Molar Flowrate",
                         "Total Mole Fraction",
                         "Temperature",
                         "Pressure"]

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):

        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp["Vap", "hydrogen"].value == \
            pytest.approx(0.077, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp["Vap", "propene"].value == \
            pytest.approx(0.077, abs=1e-4)
        assert model.props[1].phase_frac["Vap"].value == \
            pytest.approx(1.000, abs=1e-4)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.props[1].report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""Wrapper for running solvers built for Linux on Windows by using WSL2. See
https://docs.microsoft.com/en-us/windows/wsl/install-win10 for information about
setting up the WSL. Combining this utility with a batch file will allow Linux
solver binaries to be used on Windows.
"""

__author__ = "John Eslick"

import os
import re
import idaes
import click
import subprocess
from idaes.commands import cb

@cb.command(
    name="solver-wsl",
    context_settings={"ignore_unknown_options": True},
    help="Run a Linux solver on Windows using WSL")
@click.option("--distribution", required=True, help="Linux distribution")
@click.option("--user", required=True, help="User")
@click.option("--executable", required=True, help="Executable path on WSL")
@click.argument('args', nargs=-1)
def solver_wsl(distribution, user, executable, args):
    al = [None]*len(args)
    for i, a in enumerate(args):
        r = re.match(r"^([A-Za-z]):\\(.*$)", a)
        if r is not None:
            # is a Windows path (well almost certaintly)
            p = r.group(2).replace('\\', '/')
            l = r.group(1).lower()
            a = f"/mnt/{l}/{p}"
        al[i] = a
    r = subprocess.run(["wsl", "-d", distribution, "-u", user, executable] + al)
    return r.returncode



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Author: Andrew Lee
"""
import pytest
from pyomo.environ import (Block,
                           ConcreteModel,
                           Set,
                           SolverStatus,
                           TerminationCondition,
                           value)
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              fixed_variables_set,
                                              activated_constraints_set)
from idaes.core.util import get_solver

from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock)
from idaes.generic_models.properties.core.generic.generic_reaction import (
        GenericReactionParameterBlock)

from idaes.generic_models.properties.core.examples.reactions.reaction_example \
    import thermo_configuration, rxn_configuration


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.thermo_params = GenericParameterBlock(
            default=thermo_configuration)

        model.rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.thermo_params,
                     **rxn_configuration})

        rate_config = model.rxn_params.config.rate_reactions
        equil_config = model.rxn_params.config.equilibrium_reactions

        assert isinstance(model.rxn_params.rate_reaction_idx, Set)
        assert len(model.rxn_params.rate_reaction_idx) == 1
        assert "R1" in model.rxn_params.rate_reaction_idx

        assert isinstance(model.rxn_params.equilibrium_reaction_idx, Set)
        assert len(model.rxn_params.equilibrium_reaction_idx) == 1
        assert "R2" in model.rxn_params.equilibrium_reaction_idx

        assert isinstance(model.rxn_params.rate_reaction_stoichiometry, dict)
        assert len(model.rxn_params.rate_reaction_stoichiometry) == 4
        for k, v in model.rxn_params.rate_reaction_stoichiometry.items():
            if (k[1], k[2]) in rate_config[k[0]].stoichiometry.keys():
                assert v == rate_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert isinstance(model.rxn_params.equilibrium_reaction_stoichiometry,
                          dict)
        assert len(model.rxn_params.equilibrium_reaction_stoichiometry) == 4
        for k, v in model.rxn_params.equilibrium_reaction_stoichiometry.items():
            if (k[1], k[2]) in equil_config[k[0]].stoichiometry.keys():
                assert v == equil_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert isinstance(model.rxn_params.reaction_R1, Block)
        assert isinstance(model.rxn_params.reaction_R2, Block)

        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.thermo_params = GenericParameterBlock(
            default=thermo_configuration)

        model.rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.thermo_params,
                     **rxn_configuration})

        model.props = model.thermo_params.build_state_block(
                [1],
                default={"defined_state": True})

        model.rxns = model.rxn_params.build_reaction_block(
                [1],
                default={"state_block": model.props,
                         "has_equilibrium": True})

        assert_units_consistent(model)

        return model

    @pytest.mark.unit
    def test_dof(self, model):
        # Fix state
        # Note that flow of D cannot be specified due to equlibrium
        model.props[1].flow_mol_comp["A"].fix(1)
        model.props[1].flow_mol_comp["B"].fix(1)
        model.props[1].flow_mol_comp["C"].fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)

        assert degrees_of_freedom(model) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={'tol': 1e-6})
        model.rxns.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_solution(self, model):
        assert model.props[1].flow_mol_comp["D"].value == \
            pytest.approx(7.0849, rel=1e-4)

        assert model.props[1].mole_frac_phase_comp["Liq", "A"].value == \
            pytest.approx(0.09916, rel=1e-4)
        assert model.props[1].mole_frac_phase_comp["Liq", "B"].value == \
            pytest.approx(0.09916, rel=1e-4)
        assert model.props[1].mole_frac_phase_comp["Liq", "C"].value == \
            pytest.approx(0.09916, rel=1e-4)
        assert model.props[1].mole_frac_phase_comp["Liq", "D"].value == \
            pytest.approx(0.70253, rel=1e-4)

        assert value(model.rxns[1].k_rxn["R1"]) == \
            pytest.approx(0.72121, rel=1e-4)

        assert value(model.rxns[1].reaction_rate["R1"]) == pytest.approx(
            value(model.rxns[1].k_rxn["R1"] *
                  model.props[1].mole_frac_phase_comp["Liq", "A"]**1 *
                  model.props[1].mole_frac_phase_comp["Liq", "B"]**1),
            rel=1e-6)

        assert value(model.rxns[1].k_eq["R2"]) == \
            pytest.approx(71.4505, rel=1e-4)

        assert value(model.rxns[1].k_eq["R2"]) == pytest.approx(
            value(model.props[1].mole_frac_phase_comp["Liq", "D"]**1 *
                  model.props[1].mole_frac_phase_comp["Liq", "B"]**-1 *
                  model.props[1].mole_frac_phase_comp["Liq", "C"]**-1),
            rel=1e-6)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.rxns[1].report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
General tests for generic proeprties with Henry components present

Author: Andrew Lee
"""
# Import Python libraries
import pytest

# Import Pyomo components
from pyomo.environ import (ConcreteModel,
                           Set,
                           SolverStatus,
                           TerminationCondition,
                           value,
                           units as pyunits)

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock)
from idaes.core.util import get_solver

from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.eos.ideal import Ideal
from idaes.generic_models.properties.core.phase_equil import SmoothVLE
from idaes.generic_models.properties.core.phase_equil.forms import fugacity
from idaes.generic_models.properties.core.phase_equil.henry import ConstantH
import idaes.generic_models.properties.core.pure.Perrys as Perrys
import idaes.generic_models.properties.core.pure.RPP4 as RPP4
from idaes.generic_models.properties.core.phase_equil.bubble_dew import \
        IdealBubbleDew

import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)
solver = get_solver()


class TestNoHenryComps(object):
    configuration = {
        # Specifying components
        "components": {
            "A": {"type": Component,
                  "enth_mol_liq_comp": Perrys,
                  "enth_mol_ig_comp": RPP4,
                  "pressure_sat_comp": RPP4,
                  "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                  "parameter_data": {
                      "mw": 78.1136E-3,
                      "pressure_crit": 48.9e5,
                      "temperature_crit": 562.2,
                      "cp_mol_ig_comp_coeff": {'A': -3.392E1,
                                               'B': 4.739E-1,
                                               'C': -3.017E-4,
                                               'D': 7.130E-8},
                      "cp_mol_liq_comp_coeff": {'1': 1.29E2,
                                                '2': -1.7E-1,
                                                '3': 6.48E-4,
                                                '4': 0,
                                                '5': 0},
                      "enth_mol_form_liq_comp_ref": 49.0e3,
                      "enth_mol_form_vap_comp_ref": 82.9e3,
                      "pressure_sat_comp_coeff": {'A': -6.98273,
                                                  'B': 1.33213,
                                                  'C': -2.62863,
                                                  'D': -3.33399}}},
            "B": {"type": Component,
                  "enth_mol_liq_comp": Perrys,
                  "enth_mol_ig_comp": RPP4,
                  "pressure_sat_comp": RPP4,
                  "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                  "parameter_data": {
                      "mw": 92.1405E-3,
                      "pressure_crit": 41e5,
                      "temperature_crit": 591.8,
                      "cp_mol_ig_comp_coeff": {'A': -2.435E1,
                                               'B': 5.125E-1,
                                               'C': -2.765E-4,
                                               'D': 4.911E-8},
                      "cp_mol_liq_comp_coeff": {'1': 1.40E2,
                                                '2': -1.52E-1,
                                                '3': 6.95E-4,
                                                '4': 0,
                                                '5': 0},
                      "enth_mol_form_liq_comp_ref": 12.0e3,
                      "enth_mol_form_vap_comp_ref": 50.1e3,
                      "pressure_sat_comp_coeff": {'A': -7.28607,
                                                  'B': 1.38091,
                                                  'C': -2.83433,
                                                  'D': -2.79168}}}},

        # Specifying phases
        "phases":  {'Liq': {"type": LiquidPhase,
                            "equation_of_state": Ideal},
                    'Vap': {"type": VaporPhase,
                            "equation_of_state": Ideal}},

        # Declare a base units dict to save code later
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

        # Specifying state definition
        "state_definition": FTPx,
        "state_bounds": {"flow_mol": (0, 100, 1000),
                         "temperature": (273.15, 300, 450),
                         "pressure": (5e4, 1e5, 1e6)},
        "pressure_ref": 1e5,
        "temperature_ref": 300,

        # Defining phase equilibria
        "phases_in_equilibrium": [("Vap", "Liq")],
        "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
        "bubble_dew_method": IdealBubbleDew}

    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(default=self.configuration)

        model.props = model.params.build_state_block(
            [1], default={"defined_state": True})

        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["A"].fix(0.5)
        model.props[1].mole_frac_comp["B"].fix(0.5)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 2
        for i in model.params.component_list:
            assert i in ["A", "B"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 4
        for i in model.params._phase_component_set:
            assert i in [("Liq", "A"), ("Liq", "B"),
                         ("Vap", "A"), ("Vap", "B")]

        assert model.params.config.state_definition == FTPx

        assert model.params.config.state_bounds == {
                "flow_mol": (0, 100, 1000),
                "temperature": (273.15, 300, 450),
                "pressure": (5e4, 1e5, 1e6)}

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE}

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 2
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2"]

        assert model.params.phase_equilibrium_list == {
            "PE1": {"A": ("Vap", "Liq")},
            "PE2": {"B": ("Vap", "Liq")}}

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300

    @pytest.mark.unit
    def test_init_bubble_temperature(self, model):
        model.props._init_Tbub(model.props[1], pyunits.K)

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")])

    @pytest.mark.unit
    def test_init_dew_temperature(self, model):
        model.props._init_Tdew(model.props[1], pyunits.K)

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")])
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")])

    @pytest.mark.unit
    def test_init_bubble_pressure(self, model):
        model.props._init_Pbub(model.props[1], pyunits.K)

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")])

    @pytest.mark.unit
    def test_init_dew_pressure(self, model):
        model.props._init_Pdew(model.props[1], pyunits.K)

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")])

    @pytest.mark.component
    def test_solve_vle(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")])

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")])
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")])

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")])

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")])


configuration = {
    # Specifying components
    "components": {
        "A": {"type": Component,
              "enth_mol_liq_comp": Perrys,
              "enth_mol_ig_comp": RPP4,
              "pressure_sat_comp": RPP4,
              "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
              "parameter_data": {
                  "mw": 78.1136E-3,
                  "pressure_crit": 48.9e5,
                  "temperature_crit": 562.2,
                  "cp_mol_ig_comp_coeff": {'A': -3.392E1,
                                           'B': 4.739E-1,
                                           'C': -3.017E-4,
                                           'D': 7.130E-8},
                  "cp_mol_liq_comp_coeff": {'1': 1.29E2,
                                            '2': -1.7E-1,
                                            '3': 6.48E-4,
                                            '4': 0,
                                            '5': 0},
                  "enth_mol_form_liq_comp_ref": 49.0e3,
                  "enth_mol_form_vap_comp_ref": 82.9e3,
                  "pressure_sat_comp_coeff": {'A': -6.98273,
                                              'B': 1.33213,
                                              'C': -2.62863,
                                              'D': -3.33399}}},
        "B": {"type": Component,
              "enth_mol_liq_comp": Perrys,
              "enth_mol_ig_comp": RPP4,
              "pressure_sat_comp": RPP4,
              "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
              "parameter_data": {
                  "mw": 92.1405E-3,
                  "pressure_crit": 41e5,
                  "temperature_crit": 591.8,
                  "cp_mol_ig_comp_coeff": {'A': -2.435E1,
                                           'B': 5.125E-1,
                                           'C': -2.765E-4,
                                           'D': 4.911E-8},
                  "cp_mol_liq_comp_coeff": {'1': 1.40E2,
                                            '2': -1.52E-1,
                                            '3': 6.95E-4,
                                            '4': 0,
                                            '5': 0},
                  "enth_mol_form_liq_comp_ref": 12.0e3,
                  "enth_mol_form_vap_comp_ref": 50.1e3,
                  "pressure_sat_comp_coeff": {'A': -7.28607,
                                              'B': 1.38091,
                                              'C': -2.83433,
                                              'D': -2.79168}}},
        "C": {"type": Component,
              "henry_component": {"Liq": ConstantH},
              "enth_mol_liq_comp": Perrys,
              "enth_mol_ig_comp": RPP4,
              "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
              "parameter_data": {
                  "mw": 92.1405E-3,
                  "pressure_crit": 41e5,
                  "temperature_crit": 591.8,
                  "henry_ref": {"Liq": 2e5},
                  "cp_mol_ig_comp_coeff": {'A': -2.435E1,
                                           'B': 5.125E-1,
                                           'C': -2.765E-4,
                                           'D': 4.911E-8},
                  "cp_mol_liq_comp_coeff": {'1': 1.40E2,
                                            '2': -1.52E-1,
                                            '3': 6.95E-4,
                                            '4': 0,
                                            '5': 0},
                  "enth_mol_form_liq_comp_ref": 12.0e3,
                  "enth_mol_form_vap_comp_ref": 50.1e3}}},

    # Specifying phases
    "phases":  {'Liq': {"type": LiquidPhase,
                        "equation_of_state": Ideal},
                'Vap': {"type": VaporPhase,
                        "equation_of_state": Ideal}},

    # Declare a base units dict to save code later
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},

    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {"flow_mol": (0, 100, 1000),
                     "temperature": (273.15, 300, 450),
                     "pressure": (5e4, 1e5, 1e6)},
    "pressure_ref": 1e5,
    "temperature_ref": 300,

    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": IdealBubbleDew}


class TestHenryComps0(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(default=configuration)

        model.props = model.params.build_state_block(
            [1], default={"defined_state": True})

        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["A"].fix(0.5)
        model.props[1].mole_frac_comp["B"].fix(0.5)
        model.props[1].mole_frac_comp["C"].fix(1e-10)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 3
        for i in model.params.component_list:
            assert i in ["A", "B", "C"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 6
        for i in model.params._phase_component_set:
            assert i in [("Liq", "A"), ("Liq", "B"), ("Liq", "C"),
                         ("Vap", "A"), ("Vap", "B"), ("Vap", "C")]

        assert model.params.config.state_definition == FTPx

        assert model.params.config.state_bounds == {
                "flow_mol": (0, 100, 1000),
                "temperature": (273.15, 300, 450),
                "pressure": (5e4, 1e5, 1e6)}

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE}

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 3
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2", "PE3"]

        assert model.params.phase_equilibrium_list == {
            "PE1": {"A": ("Vap", "Liq")},
            "PE2": {"B": ("Vap", "Liq")},
            "PE3": {"C": ("Vap", "Liq")}}

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300
        assert model.params.C.henry_ref_Liq.value == 2e5

    @pytest.mark.unit
    def test_init_bubble_temperature(self, model):
        model.props._init_Tbub(model.props[1], pyunits.K)

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")])

    @pytest.mark.unit
    def test_init_dew_temperature(self, model):
        model.props._init_Tdew(model.props[1], pyunits.K)

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")])
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")])

    @pytest.mark.unit
    def test_init_bubble_pressure(self, model):
        model.props._init_Pbub(model.props[1], pyunits.K)

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")])

    @pytest.mark.unit
    def test_init_dew_pressure(self, model):
        model.props._init_Pdew(model.props[1], pyunits.K)

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")])

    @pytest.mark.component
    def test_solve_vle(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert pytest.approx(365.35, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7137, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2863, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")])

        assert pytest.approx(372.02, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")])
        assert pytest.approx(0.2909, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7091, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")])

        assert pytest.approx(109479, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")])

        assert pytest.approx(89820, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")])
        assert pytest.approx(0.2881, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.7119, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")])


class TestHenryComps(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(default=configuration)

        model.props = model.params.build_state_block(
            [1], default={"defined_state": True})

        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["A"].fix(0.45)
        model.props[1].mole_frac_comp["B"].fix(0.45)
        model.props[1].mole_frac_comp["C"].fix(0.1)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 3
        for i in model.params.component_list:
            assert i in ["A", "B", "C"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 6
        for i in model.params._phase_component_set:
            assert i in [("Liq", "A"), ("Liq", "B"), ("Liq", "C"),
                         ("Vap", "A"), ("Vap", "B"), ("Vap", "C")]

        assert model.params.config.state_definition == FTPx

        assert model.params.config.state_bounds == {
                "flow_mol": (0, 100, 1000),
                "temperature": (273.15, 300, 450),
                "pressure": (5e4, 1e5, 1e6)}

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE}

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 3
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2", "PE3"]

        assert model.params.phase_equilibrium_list == {
            "PE1": {"A": ("Vap", "Liq")},
            "PE2": {"B": ("Vap", "Liq")},
            "PE3": {"C": ("Vap", "Liq")}}

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300
        assert model.params.C.henry_ref_Liq.value == 2e5

    @pytest.mark.unit
    def test_init_bubble_temperature(self, model):
        model.props._init_Tbub(model.props[1], pyunits.K)

        assert pytest.approx(361.50, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")])
        assert pytest.approx(0.5750, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2276, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")])
        assert pytest.approx(0.1974, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "C")])

    @pytest.mark.unit
    def test_init_dew_temperature(self, model):
        model.props._init_Tdew(model.props[1], pyunits.K)

        assert pytest.approx(370.23, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")])
        assert pytest.approx(0.2750, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.6744, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")])
        assert pytest.approx(0.0506, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "C")])

    @pytest.mark.unit
    def test_init_bubble_pressure(self, model):
        model.props._init_Pbub(model.props[1], pyunits.K)

        assert pytest.approx(118531, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")])
        assert pytest.approx(0.5918, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2395, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")])
        assert pytest.approx(0.1687, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "C")])

    @pytest.mark.unit
    def test_init_dew_pressure(self, model):
        model.props._init_Pdew(model.props[1], pyunits.K)

        assert pytest.approx(95056, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")])
        assert pytest.approx(0.2744, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.6780, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")])
        assert pytest.approx(0.0476, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "C")])

    @pytest.mark.component
    def test_solve_vle(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert pytest.approx(361.50, abs=0.01) == value(
            model.props[1].temperature_bubble[("Vap", "Liq")])
        assert pytest.approx(0.5750, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2276, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "B")])
        assert pytest.approx(0.1974, abs=1e-4) == value(
            model.props[1]._mole_frac_tbub[("Vap", "Liq", "C")])

        assert pytest.approx(370.23, abs=0.01) == value(
            model.props[1].temperature_dew[("Vap", "Liq")])
        assert pytest.approx(0.2750, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.6744, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "B")])
        assert pytest.approx(0.0506, abs=1e-4) == value(
            model.props[1]._mole_frac_tdew[("Vap", "Liq", "C")])

        assert pytest.approx(118531, abs=1) == value(
            model.props[1].pressure_bubble[("Vap", "Liq")])
        assert pytest.approx(0.5918, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "A")])
        assert pytest.approx(0.2395, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "B")])
        assert pytest.approx(0.1687, abs=1e-4) == value(
            model.props[1]._mole_frac_pbub[("Vap", "Liq", "C")])

        assert pytest.approx(95056, abs=1) == value(
            model.props[1].pressure_dew[("Vap", "Liq")])
        assert pytest.approx(0.2744, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "A")])
        assert pytest.approx(0.6780, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "B")])
        assert pytest.approx(0.0476, abs=1e-4) == value(
            model.props[1]._mole_frac_pdew[("Vap", "Liq", "C")])
